import hub
import numpy as np
from typing import Any, Dict, List, Optional, Tuple
from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.tiling.sample_tiles import SampleTiles


class TileEncoder(HubMemoryObject):
    def __init__(self, entries=None, version=None):
        self.is_dirty = False
        self.entries: Dict[int, Tuple[Tuple[int, ...], Tuple[int, ...]]] = entries or {}
        self.version = version or hub.__version__

    def register_sample(self, sample: SampleTiles, idx: int):
        """Registers a new tiled sample into the encoder.

        Args:
            sample: The sample to be registered.
            idx: The global sample index.
        """
        if sample.registered:
            return
        ss: Tuple[int, ...] = sample.sample_shape  # type: ignore
        ts: Tuple[int, ...] = sample.tile_shape
        self.entries[idx] = (ss, ts)
        sample.registered = True
        self.is_dirty = True

    def __delitem__(self, global_sample_index: int):
        del self.entries[global_sample_index]
        keys = sorted(self.entries.keys())
        for k in keys:
            if k > global_sample_index:
                self.entries[k - 1] = self.entries[k]
                del self.entries[k]
        self.is_dirty = True

    def __getitem__(self, global_sample_index: int):
        return self.entries[global_sample_index]

    def __contains__(self, global_sample_index: int):
        """Returns whether the index is present in the tile encoder. Useful for checking if a given sample is tiled."""
        return global_sample_index in self.entries

    def get_tile_shape(self, global_sample_index: int):
        return tuple(self[global_sample_index][1])

    def get_sample_shape(self, global_sample_index: int):
        return tuple(self[global_sample_index][0])

    def get_tile_layout_shape(self, global_sample_index: int) -> Tuple[int, ...]:
        """If you were to lay the tiles out in a grid, the tile layout shape would be the shape
        of the grid.

        Example:
            Sample shape:               (1000, 500)
            Tile shape:                 (10, 10)
            Output tile layout shape:   (100, 50)
        """

        tile_meta = self[global_sample_index]
        tile_shape = tile_meta[1]
        sample_shape = tile_meta[0]

        if len(tile_shape) != len(sample_shape):
            raise ValueError(
                "Tile shape and sample shape must have the same number of dimensions."
            )

        layout = [
            np.ceil(sample_shape_dim / tile_shape_dim)
            for tile_shape_dim, sample_shape_dim in zip(tile_shape, sample_shape)
        ]
        return tuple(int(x) for x in layout)

    @property
    def nbytes(self):
        len_version = 1 + len(self.version)
        entries = self.entries
        num_entries = len(entries)
        if num_entries == 0:
            return len_version + 8
        value = next(iter(entries.values()))
        num_dimensions = len(value[0])
        return len_version + 16 + (num_entries * (8 * (1 + num_dimensions * 2)))

    def tobytes(self) -> memoryview:
        """Serialize entries dict into bytes

        Corresponding to every key, there is a tuple of 2 shapes which are tuples.
        Example of entries is:-
        {
            0: (
                (1000, 500, 100, 3),
                (10, 10, 10, 3)
            ),
            7: (
                (3000, 600, 100, 4),
                (50, 50, 50, 4)
            )
        }
        All the tuples inside will have the same number of dimensions.
        The format stores the following information in order:
        1. First byte tells the length of the version string, call this v.
        2. The next v bytes are the version string.
        3. The next 8 bytes are the number of entries call this n.
        4. The next 8 bytes are the number of dimensions in the shapes, call this d.
        5. The next n * (8 * (1 + d + d)) bytes are the entries in the format:
            a. The first 8 bytes are the key.
            b. The next d * 8 bytes are the first shape in value.
            c. The next d * 8 bytes are the second shape in value.
        """
        entries = self.entries
        num_entries = len(entries)

        data = bytearray(self.nbytes)
        ofs = 0

        # store version length
        data[ofs] = len(self.version)
        ofs += 1

        # store version string
        data[ofs : ofs + len(self.version)] = self.version.encode("ascii")
        ofs += len(self.version)

        # store the number of entries
        data[ofs : ofs + 8] = num_entries.to_bytes(8, byteorder="little")
        ofs += 8

        if num_entries == 0:
            return memoryview(data)

        value = next(iter(entries.values()))
        num_dimensions = len(value[0])

        # store the number of dimensions
        data[ofs : ofs + 8] = num_dimensions.to_bytes(8, byteorder="little")
        ofs += 8

        # store the entries
        for key, value in entries.items():
            # store the key
            data[ofs : ofs + 8] = key.to_bytes(8, byteorder="little")
            ofs += 8

            # store the first shape
            first_shape = value[0]
            for dimension in first_shape:
                data[ofs : ofs + 8] = dimension.to_bytes(8, byteorder="little")
                ofs += 8

            # store the second shape
            second_shape = value[1]
            for dimension in second_shape:
                data[ofs : ofs + 8] = dimension.to_bytes(8, byteorder="little")
                ofs += 8

        return memoryview(data)

    @classmethod
    def frombuffer(cls, data: bytes):
        """Deserialize bytes into entries dict"""
        try:
            ofs = 0
            # Get version length
            version_length = data[ofs]
            ofs += 1

            # Get version string
            version = str(data[ofs : ofs + version_length], "ascii")
            ofs += version_length
            check_version(version)
            entries = parse_tile_encoder_entries(data, ofs, "little")
            enc = cls(entries, version=version)
        except Exception:
            # backwards compatibility
            entries = parse_tile_encoder_entries(data, 0, "big")
            enc = cls(entries)
        enc.is_dirty = False
        return enc

    def __getstate__(self) -> Dict[str, Any]:
        return {"entries": self.entries, "version": self.version}

    def __setstate__(self, state: Dict[str, Any]):
        self.entries = state["entries"]
        self.version = state["version"]


def parse_tile_encoder_entries(data, ofs: int, byteorder: str) -> Dict:
    # Get the number of entries
    num_entries = int.from_bytes(data[ofs : ofs + 8], byteorder=byteorder)  # type: ignore
    ofs += 8
    if num_entries == 0:
        return {}

    # Get the number of dimensions of the tuples
    num_dim = int.from_bytes(data[ofs : ofs + 8], byteorder=byteorder)  # type: ignore
    ofs += 8

    # Get the entries
    entries = {}
    for _ in range(num_entries):
        # Get the key
        key = int.from_bytes(data[ofs : ofs + 8], byteorder=byteorder)  # type: ignore
        ofs += 8

        first_shape: List[int] = []
        second_shape: List[int] = []

        for shp in [first_shape, second_shape]:
            for _ in range(num_dim):
                shp.append(int.from_bytes(data[ofs : ofs + 8], byteorder=byteorder))  # type: ignore
                ofs += 8

        # Add the entry to the dict
        entries[key] = (tuple(first_shape), tuple(second_shape))
    return entries


def check_version(version):
    if len(version) < 5 or len(version) > 7:
        raise ValueError("Invalid version length")
    if version.count(".") < 2 or version[0] != "2":
        raise ValueError("Invalid version format")



from hub.util.exceptions import LockedException
import numpy as np
import pytest
import hub
import uuid
import time
import warnings
from hub.tests.dataset_fixtures import enabled_cloud_dataset_generators


_counter = 0


class VM(object):
    """
    Emulates a different machine
    """

    def __init__(self):
        global _counter
        self.id = _counter
        _counter += 1

    def __enter__(self):
        self._getnode = uuid.getnode
        uuid.getnode = lambda: self.id
        self._locks = hub.core.lock._LOCKS.copy()
        hub.core.lock._LOCKS.clear()

    def __exit__(self, *args, **kwargs):
        uuid.getnode = self._getnode
        hub.core.lock._LOCKS.update(self._locks)


@enabled_cloud_dataset_generators
def test_dataset_locking(ds_generator):
    ds = ds_generator()
    ds.create_tensor("x")
    arr = np.random.random((32, 32))
    ds.x.append(arr)

    with VM():
        # Make sure read only warning is raised
        with pytest.warns(UserWarning):
            ds = ds_generator()
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
        assert ds.read_only == True
        with pytest.raises(LockedException):
            ds.read_only = False
        # Raise error if user explicitly asks for write access
        with pytest.raises(LockedException):
            ds = ds_generator(read_only=False)
        # No warnings if user requests read only mode
        with warnings.catch_warnings(record=True) as ws:
            ds = ds_generator(read_only=True)
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
        assert not ws

        DATASET_LOCK_VALIDITY = hub.constants.DATASET_LOCK_VALIDITY
        # Temporarily set validity to 1 second so we dont have to wait too long.
        hub.constants.DATASET_LOCK_VALIDITY = 1
        # Wait for lock to expire.
        time.sleep(1.1)

        try:
            ds = ds_generator()
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
            assert ds.read_only == False
        finally:
            hub.constants.DATASET_LOCK_VALIDITY = DATASET_LOCK_VALIDITY


@enabled_cloud_dataset_generators
def test_vc_locking(ds_generator):
    ds = ds_generator()
    ds.create_tensor("x")
    arr = np.random.random((32, 32))
    ds.x.append(arr)
    ds.commit()
    ds.checkout("branch", create=True)
    with VM():
        with warnings.catch_warnings(record=True) as ws:
            ds = ds_generator()
        np.testing.assert_array_equal(arr, ds.x[0].numpy())
        assert not ws, str(ws[0])


def test_lock_thread_leaking(s3_ds_generator):
    locks = hub.core.lock._LOCKS
    refs = hub.core.lock._REFS
    nlocks_previous = len(locks)

    def nlocks():
        assert len(locks) == len(refs)
        return len(locks) - nlocks_previous

    ds = s3_ds_generator()
    ds.create_tensor("a")
    assert nlocks() == 1

    ds.__del__()  # Note: investigate why this doesnt happen automatically. (cyclic refs?)
    del ds
    assert nlocks() == 0

    ds = s3_ds_generator()
    ds.create_tensor("x")
    ds.x.extend(np.random.random((2, 32)))
    views = []
    for i in range(32):
        views.append(ds[i : i + 1])

    ds.__del__()
    del ds

    assert nlocks() == 1  # 1 because views

    views.pop()
    assert nlocks() == 1  # deleting 1 view doesn't release locks

    del views
    assert nlocks() == 0  # 0 because dataset and all views deleted



# type: ignore

from typing import Tuple, Optional
from hub.htype import htype as HTYPE, HTYPE_CONFIGURATIONS
from hub.util.exceptions import TensorMetaInvalidHtype


def parse_complex_htype(htype: Optional[str]) -> Tuple[bool, bool, str]:
    is_sequence = False
    is_link = False

    if not htype:
        htype = HTYPE.DEFAULT

    elif htype.startswith("sequence"):
        is_sequence, is_link, htype = parse_sequence_start(htype)

    elif htype.startswith("link"):
        is_sequence, is_link, htype = parse_link_start(htype)

    if "[" in htype or "]" in htype:
        raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))

    if is_link and htype == HTYPE.DEFAULT:
        if is_sequence:
            raise ValueError(
                "Can't create a linked tensor with a generic htype, you need to specify htype, for example sequence[link[image]] or link[sequence[image]]"
            )
        raise ValueError(
            "Can't create a linked tensor with a generic htype, you need to specify htype, for example link[image]"
        )
    return is_sequence, is_link, htype


def parse_sequence_start(htype):
    if htype == "sequence":
        return True, False, HTYPE.DEFAULT
    if htype[len("sequence")] != "[" or htype[-1] != "]":
        raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
    htype = htype.split("[", 1)[1][:-1]
    if not htype:
        return True, False, HTYPE.DEFAULT
    if htype.startswith("link"):
        if htype == "link":
            return True, True, HTYPE.DEFAULT
        if htype[len("link")] != "[" or htype[-1] != "]":
            raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
        htype = htype.split("[", 1)[1][:-1]
        if not htype:
            return True, True, HTYPE.DEFAULT
        if "[" in htype or "]" in htype:
            raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
        return True, True, htype
    return True, False, htype


def parse_link_start(htype):
    if htype == "link":
        return False, True, HTYPE.DEFAULT
    if htype[len("link")] != "[" or htype[-1] != "]":
        raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
    htype = htype.split("[", 1)[1][:-1]
    if not htype:
        return False, True, HTYPE.DEFAULT
    if htype.startswith("sequence"):
        if htype == "sequence":
            return True, True, HTYPE.DEFAULT
        if htype[len("sequence")] != "[" or htype[-1] != "]":
            raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
        htype = htype.split("[", 1)[1][:-1]
        if not htype:
            return True, True, HTYPE.DEFAULT
        if "[" in htype or "]" in htype:
            raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS))
        return True, True, htype
    return False, True, htype



from math import floor
from hub.tests.common import requires_linux
from hub.experimental import query


@requires_linux
def test_query(local_ds):
    with local_ds as ds:
        ds.create_tensor("label")
        for i in range(100):
            ds.label.append(floor(i / 20))

    dsv = query(local_ds, "SELECT * WHERE CONTAINS(label, 2)")
    assert len(dsv) == 20
    for i in range(20):
        assert dsv.label[i].numpy() == 2



from collections import defaultdict, OrderedDict
from typing import Any, DefaultDict, Dict, List, Optional, Set, Tuple
from hub.core.meta.dataset_meta import DatasetMeta
from hub.core.version_control.commit_diff import CommitDiff
from hub.core.version_control.commit_node import CommitNode  # type: ignore
from hub.core.storage import LRUCache
from hub.core.version_control.dataset_diff import DatasetDiff
from hub.util.keys import (
    get_dataset_diff_key,
    get_dataset_meta_key,
    get_tensor_commit_diff_key,
)


def get_changes_and_messages(
    version_state, storage, id_1, id_2
) -> Tuple[
    dict, Optional[dict], dict, Optional[dict], Optional[str], str, Optional[str]
]:
    if id_1 is None and id_2 is None:
        return get_changes_and_messages_compared_to_prev(version_state, storage)
    return get_changes_and_message_2_ids(version_state, storage, id_1, id_2)


def get_changes_and_messages_compared_to_prev(
    version_state, storage
) -> Tuple[dict, None, dict, None, None, str, None]:
    commit_node = version_state["commit_node"]
    commit_id = commit_node.commit_id
    head = commit_node.is_head_node

    tensor_changes: Dict[str, Dict] = defaultdict(dict)
    ds_changes: Dict[str, Any] = {}
    s = "HEAD" if head else f"{commit_id} (current commit)"
    msg_1 = f"Diff in {s} relative to the previous commit:\n"
    get_tensor_changes_for_id(commit_id, storage, tensor_changes, ds_changes)
    get_dataset_changes_for_id(commit_id, storage, ds_changes)

    combine_data_deleted(tensor_changes)
    filter_cleared(tensor_changes)
    filter_data_updated(tensor_changes)
    filter_renamed_diff(ds_changes)
    remove_empty_changes(tensor_changes)

    # Order: ds_changes_1, ds_changes_2, tensor_changes_1, tensor_changes_2, msg_0, msg_1, msg_2
    return ds_changes, None, tensor_changes, None, None, msg_1, None


def get_changes_and_message_2_ids(
    version_state, storage, id_1, id_2
) -> Tuple[dict, dict, dict, dict, str, str, str]:
    commit_node = version_state["commit_node"]
    if id_1 is None:
        raise ValueError("Can't specify id_2 without specifying id_1")
    msg_0 = "The 2 diffs are calculated relative to the most recent common ancestor (%s) of the "
    if id_2 is None:
        msg_0 += "current state and the commit passed."
        id_2 = id_1
        id_1 = commit_node.commit_id
        head = commit_node.is_head_node
        msg_1 = "Diff in HEAD:\n" if head else f"Diff in {id_1} (current commit):\n"
        msg_2 = f"Diff in {id_2} (target id):\n"
    else:
        msg_0 += "two commits passed."
        msg_1 = f"Diff in {id_1} (target id 1):\n"
        msg_2 = f"Diff in {id_2} (target id 2):\n"

    ret = compare_commits(id_1, id_2, version_state, storage)
    ds_changes_1, ds_changes_2, tensor_changes_1, tensor_changes_2, lca_id = ret
    msg_0 %= lca_id
    remove_empty_changes(tensor_changes_1)
    remove_empty_changes(tensor_changes_2)
    return (
        ds_changes_1,
        ds_changes_2,
        tensor_changes_1,
        tensor_changes_2,
        msg_0,
        msg_1,
        msg_2,
    )


def compare_commits(
    id_1: str, id_2: str, version_state: Dict[str, Any], storage: LRUCache
) -> Tuple[dict, dict, dict, dict, str]:
    """Compares two commits and returns the differences.

    Args:
        id_1 (str): The first commit_id or branch name.
        id_2 (str): The second commit_id or branch name.
        version_state (dict): The version state.
        storage (LRUCache): The underlying storage of the dataset.

    Returns:
        Tuple[dict, dict, dict, dict, str]: The changes made in the first commit and second commit respectively, followed by lca_id.
    """
    id_1 = sanitize_commit(id_1, version_state)
    id_2 = sanitize_commit(id_2, version_state)
    mp = version_state["commit_node_map"]
    commit_node_1: CommitNode = mp[id_1]
    commit_node_2: CommitNode = mp[id_2]
    lca_id = get_lowest_common_ancestor(commit_node_1, commit_node_2)
    lca_node: CommitNode = mp[lca_id]

    tensor_changes_1: Dict[str, Dict] = defaultdict(dict)
    tensor_changes_2: Dict[str, Dict] = defaultdict(dict)
    dataset_changes_1: Dict[str, Any] = {}
    dataset_changes_2: Dict[str, Any] = {}

    for commit_node, tensor_changes, dataset_changes in [
        (commit_node_1, tensor_changes_1, dataset_changes_1),
        (commit_node_2, tensor_changes_2, dataset_changes_2),
    ]:
        while commit_node.commit_id != lca_node.commit_id:
            commit_id = commit_node.commit_id
            get_tensor_changes_for_id(
                commit_id, storage, tensor_changes, dataset_changes
            )
            get_dataset_changes_for_id(commit_id, storage, dataset_changes)
            commit_node = commit_node.parent  # type: ignore

        combine_data_deleted(tensor_changes)
        filter_cleared(tensor_changes)
        filter_data_updated(tensor_changes)
        filter_renamed_diff(dataset_changes)
        remove_empty_changes(tensor_changes)

    return (
        dataset_changes_1,
        dataset_changes_2,
        tensor_changes_1,
        tensor_changes_2,
        lca_id,
    )


def sanitize_commit(id: str, version_state: Dict[str, Any]) -> str:
    """Checks the id.
    If it's a valid commit_id, it is returned.
    If it's a branch name, the commit_id of the branch's head is returned.
    Otherwise a ValueError is raised.
    """
    if id in version_state["commit_node_map"]:
        return id
    elif id in version_state["branch_commit_map"]:
        return version_state["branch_commit_map"][id]
    raise KeyError(f"The commit/branch {id} does not exist in the dataset.")


def get_lowest_common_ancestor(p: CommitNode, q: CommitNode):
    """Returns the lowest common ancestor of two commits."""
    if p == q:
        return p.commit_id

    p_family = []
    q_family = set()

    while p:
        p_family.append(p.commit_id)
        p = p.parent  # type: ignore

    while q:
        q_family.add(q.commit_id)
        q = q.parent  # type: ignore
    for id in p_family:
        if id in q_family:
            return id


def get_all_changes_string(
    ds_changes_1, ds_changes_2, tensor_changes_1, tensor_changes_2, msg_0, msg_1, msg_2
):
    """Returns a string with all changes."""
    all_changes = ["\n## Hub Diff"]
    if msg_0:
        all_changes.append(msg_0)

    separator = "-" * 120
    if tensor_changes_1 is not None:
        changes1_str = get_changes_str(ds_changes_1, tensor_changes_1, msg_1, separator)
        all_changes.append(changes1_str)
    if tensor_changes_2 is not None:
        changes2_str = get_changes_str(ds_changes_2, tensor_changes_2, msg_2, separator)
        all_changes.append(changes2_str)
    all_changes.append(separator)
    return "\n".join(all_changes)


def get_changes_str(ds_changes, tensor_changes: Dict, message: str, separator: str):
    """Returns a string with changes made."""
    all_changes = [separator, message]
    if ds_changes.get("info_updated", False):
        all_changes.append("- Updated dataset info \n")
    if ds_changes.get("deleted"):
        for name in ds_changes["deleted"]:
            all_changes.append(f"- Deleted:\t{name}")
    if ds_changes.get("renamed"):
        for old, new in ds_changes["renamed"].items():
            all_changes.append(f"- Renamed:\t{old} -> {new}")
    if len(all_changes) > 2:
        all_changes.append("\n")

    tensors = sorted(tensor_changes.keys())
    for tensor in tensors:
        change = tensor_changes[tensor]
        created = change.get("created", False)
        cleared = change.get("cleared", False)
        data_added = change.get("data_added", [0, 0])
        data_added_str = convert_adds_to_string(data_added)
        data_updated = change.get("data_updated", set())
        info_updated = change.get("info_updated", False)

        data_deleted = change.get("data_deleted", set())
        all_changes.append(tensor)
        if created:
            all_changes.append("* Created tensor")

        if cleared:
            all_changes.append("* Cleared tensor")

        if data_added_str:
            all_changes.append(data_added_str)

        if data_updated:
            output = convert_updates_deletes_to_string(data_updated, "Updated")
            all_changes.append(output)

        if data_deleted:
            output = convert_updates_deletes_to_string(data_deleted, "Deleted")
            all_changes.append(output)

        if info_updated:
            all_changes.append("* Updated tensor info")
        all_changes.append("")
    if len(all_changes) == 2:
        all_changes.append("No changes were made.")
    return "\n".join(all_changes)


def has_change(change: Dict):
    created = change.get("created", False)
    cleared = change.get("cleared")
    data_added = change.get("data_added", [0, 0])
    num_samples_added = data_added[1] - data_added[0]
    data_updated = change.get("data_updated", set())
    info_updated = change.get("info_updated", False)
    data_deleted = change.get("data_deleted", set())
    return (
        created
        or cleared
        or num_samples_added > 0
        or data_updated
        or info_updated
        or data_deleted
    )


def get_dataset_changes_for_id(
    commit_id: str,
    storage: LRUCache,
    dataset_changes,
):
    """Returns the changes made in the dataset for a commit."""

    dataset_diff_key = get_dataset_diff_key(commit_id)
    try:
        dataset_diff = storage.get_hub_object(dataset_diff_key, DatasetDiff)
        dataset_changes["info_updated"] = (
            dataset_changes.get("info_updated") or dataset_diff.info_updated
        )

        renamed = dataset_changes.get("renamed")
        deleted = dataset_changes.get("deleted")
        done = []

        merge_renamed = OrderedDict()
        for old, new in dataset_diff.renamed.items():
            if deleted and new in deleted and new not in done:
                deleted[deleted.index(new)] = old
                done.append(old)
                continue
            if renamed and renamed.get(new):
                merge_renamed[old] = renamed[new]
                renamed.pop(new)
            else:
                merge_renamed[old] = new

        try:
            dataset_changes["renamed"].update(merge_renamed)
        except KeyError:
            dataset_changes["renamed"] = merge_renamed

        if dataset_changes.get("deleted"):
            dataset_changes["deleted"].extend(dataset_diff.deleted)
        else:
            dataset_changes["deleted"] = dataset_diff.deleted.copy()
    except KeyError:
        pass


def get_tensor_changes_for_id(
    commit_id: str,
    storage: LRUCache,
    tensor_changes: Dict[str, Dict],
    dataset_changes,
):
    """Identifies the changes made in the given commit_id and updates them in the changes dict."""
    meta_key = get_dataset_meta_key(commit_id)
    meta = storage.get_hub_object(meta_key, DatasetMeta)
    tensors = meta.visible_tensors

    for tensor in tensors:
        key = meta.tensor_names[tensor]
        try:
            commit_diff: CommitDiff
            commit_diff_key = get_tensor_commit_diff_key(key, commit_id)
            commit_diff = storage.get_hub_object(commit_diff_key, CommitDiff)
            renamed = dataset_changes.get("renamed")
            deleted = dataset_changes.get("deleted")

            if deleted and tensor in deleted:
                if commit_diff.created:
                    deleted.remove(tensor)
                continue

            if renamed:
                try:
                    new_name = renamed[tensor]
                    if commit_diff.created:
                        renamed.pop(tensor, None)
                    tensor = new_name
                except KeyError:
                    pass

            change = tensor_changes[tensor]

            change["created"] = change.get("created") or commit_diff.created
            # ignore older diffs if tensor was cleared
            if change.get("cleared"):
                continue

            change["cleared"] = change.get("cleared") or commit_diff.cleared

            change["info_updated"] = (
                change.get("info_updated") or commit_diff.info_updated
            )

            # this means that the data was transformed inplace in a newer commit, so we can ignore older diffs
            if change.get("data_transformed_in_place", False):
                continue

            if "data_added" not in change:
                change["data_added"] = commit_diff.data_added.copy()
            else:
                change["data_added"][0] = (
                    change["data_added"][0] - commit_diff.num_samples_added
                )

            if "data_updated" not in change:
                change["data_updated"] = commit_diff.data_updated.copy()
            else:
                change["data_updated"].update(commit_diff.data_updated)
            change["data_transformed_in_place"] = (
                change.get("data_transformed_in_place") or commit_diff.data_transformed
            )

            if "data_deleted_list" not in change:
                change["data_deleted_list"] = [commit_diff.data_deleted]
            else:
                change["data_deleted_list"].append(commit_diff.data_deleted)

        except KeyError:
            pass


def combine_data_deleted(changes: Dict[str, Dict]):
    """Combines the data deleted list into a single set."""
    for change in changes.values():
        data_deleted: Set[int] = set()
        data_deleted_list = change.pop("data_deleted_list", [])
        for deleted_list in reversed(data_deleted_list):
            for index in deleted_list:
                offset = sum(i < index for i in data_deleted)
                data_deleted.add(index + offset)
        change["data_deleted"] = data_deleted


def filter_data_updated(changes: Dict[str, Dict]):
    """Removes the intersection of data added and data updated from data updated."""
    for change in changes.values():
        deleted_data = set(change.get("data_deleted", set()))
        # only show the elements in data_updated that are not in data_added
        data_added_range = range(change["data_added"][0], change["data_added"][1] + 1)
        upd = {
            data
            for data in change["data_updated"]
            if data not in data_added_range and data not in deleted_data
        }
        change["data_updated"] = upd


def filter_renamed_diff(dataset_changes):
    """Remove deleted tensors and tensors renamed to same name from diff"""
    rm = []
    renamed = dataset_changes.get("renamed")
    deleted = dataset_changes.get("deleted")
    if renamed:
        for old_name, new_name in renamed.items():
            if old_name == new_name:
                rm.append(old_name)

        for name in rm:
            renamed.pop(name)


def filter_cleared(changes: Dict[str, Dict]):
    """Removes cleared flag if created flag is true."""
    for change in changes.values():
        if change["created"] and change["cleared"]:
            change["cleared"] = False


def compress_into_range_intervals(indexes: Set[int]) -> List[Tuple[int, int]]:
    """Compresses the indexes into range intervals.
    Examples:
        compress_into_range_intervals({1, 2, 3, 4, 5, 6, 7, 8, 9, 10})
        >> [(1, 10)]
        compress_into_range_intervals({1,2,3,5,6,8,10})
        >> [(1, 3), (5, 6), (8,8), (10,10)]

    Args:
        indexes (Set[int]): The indexes to compress.

    Returns:
        List[Tuple[int, int]]: The compressed range intervals.
    """

    if not indexes:
        return []

    sorted_indexes = sorted(indexes)
    compressed_indexes: List[Tuple[int, int]] = []
    start = sorted_indexes[0]
    end = sorted_indexes[0]
    for index in sorted_indexes[1:]:
        if index != end + 1:
            compressed_indexes.append((start, end))
            start = index
        end = index
    compressed_indexes.append((start, end))
    return compressed_indexes


def range_interval_list_to_string(range_intervals: List[Tuple[int, int]]) -> str:
    """Converts the range intervals to a string.

    Examples:
        range_interval_list_to_string([(1, 10)])
        >>"1-10"
        range_interval_list_to_string([(1, 3), (5, 6), (8, 8) (10, 10)])
        >>"1-3, 5-6, 8, 10"

    Args:
        range_intervals (List[Tuple[int, int]]): The range intervals to convert.

    Returns:
        str: The string representation of the range intervals.
    """
    if not range_intervals:
        return ""
    output = ""
    for start, end in range_intervals:
        if start == end:
            output += f"{start}, "
        else:
            output += f"{start}-{end}, "
    return output[:-2]


def convert_updates_deletes_to_string(indexes: Set[int], operation: str) -> str:
    range_intervals = compress_into_range_intervals(indexes)
    output = range_interval_list_to_string(range_intervals)

    num_samples = len(indexes)
    sample_string = "sample" if num_samples == 1 else "samples"
    return f"* {operation} {num_samples} {sample_string}: [{output}]"


def convert_adds_to_string(index_range: List[int]) -> str:
    num_samples = index_range[1] - index_range[0]
    if num_samples == 0:
        return ""
    sample_string = "sample" if num_samples == 1 else "samples"
    return f"* Added {num_samples} {sample_string}: [{index_range[0]}-{index_range[1]}]"


def remove_empty_changes(changes: Dict):
    if not changes:
        return
    tensors = list(changes.keys())
    for tensor in tensors:
        change = changes[tensor]
        if not has_change(change):
            del changes[tensor]



import hub
import time
import uuid
import struct
import atexit
import threading

from typing import Tuple, Dict, Callable, Optional, Set
from collections import defaultdict
from hub.util.exceptions import LockedException
from hub.util.keys import get_dataset_lock_key
from hub.util.remove_cache import get_base_storage
from hub.util.path import get_path_from_storage
from hub.util.threading import terminate_thread
from hub.core.storage import StorageProvider
from hub.constants import FIRST_COMMIT_ID
from hub.client.utils import get_user_name


def _get_lock_bytes(username: Optional[str] = None) -> bytes:
    byts = uuid.getnode().to_bytes(6, "little") + struct.pack("d", time.time())
    ## TODO Uncomment lines below once this version has propogated through the userbase.
    # if username:
    #     byts += username.encode("utf-8")
    return byts


def _parse_lock_bytes(byts) -> Tuple[int, int, str]:
    byts = memoryview(byts)
    nodeid = int.from_bytes(byts[:6], "little")
    timestamp = struct.unpack("d", byts[6:14])[0]
    username = str(byts[14:], "utf-8")
    return nodeid, timestamp, username


class Lock(object):
    def __init__(self, storage: StorageProvider, path: str):
        self.storage = storage
        self.path = path
        username = get_user_name()
        if username == "public":
            self.username = None
        else:
            self.username = username

    def _write_lock(self):
        storage = self.storage
        try:
            read_only = storage.read_only
            storage.disable_readonly()
            storage[self.path] = _get_lock_bytes(self.username)
        finally:
            if read_only:
                storage.enable_readonly()

    def acquire(self, timeout=10, force=False):
        storage = self.storage
        path = self.path
        try:
            nodeid, timestamp, _ = _parse_lock_bytes(storage[path])
        except KeyError:
            return self._write_lock()
        if nodeid == uuid.getnode():
            return self._write_lock()
        while path in storage:
            if time.time() - timestamp >= timeout:
                if force:
                    return self._write_lock()
                else:
                    raise LockedException()
            time.sleep(1)

    def release(self):
        storage = self.storage
        try:
            read_only = storage.read_only
            storage.disable_readonly()
            del storage[self.path]
        except Exception:
            pass
        finally:
            if read_only:
                storage.enable_readonly()

    def __enter__(self):
        self.acquire()

    def __exit__(self, *args, **kwargs):
        self.release()


class PersistentLock(Lock):
    """Locks a StorageProvider instance to avoid concurrent writes from multiple machines.

    Example:
        From machine 1:
        s3 = hub.core.storage.S3Provider(S3_URL)
        lock = hub.core.lock.Lock(s3)  # Works

        From machine 2:
        s3 = hub.core.storage.S3Provider(S3_URL)
        lock = hub.core.lock.Lock(s3)  # Raises LockedException

        The lock is updated every 2 mins by an internal thread. The lock is valid for 5 mins after the last update.

    Args:
        storage (StorageProvider): The storage provder to be locked.
        lock_lost_callback (Callable, optional): Called if the lock is lost after acquiring.

    Raises:
        LockedException: If the storage is already locked by a different machine.
    """

    def __init__(
        self,
        storage: StorageProvider,
        path: Optional[str] = None,
        lock_lost_callback: Optional[Callable] = None,
    ):
        self.storage = storage
        self.path = get_dataset_lock_key() if path is None else path
        self.lock_lost_callback = lock_lost_callback
        self.acquired = False
        self._thread_lock = threading.Lock()
        self._previous_update_timestamp = None
        username = get_user_name()
        if username == "public":
            self.username = None
        else:
            self.username = username
        self.acquire()
        atexit.register(self.release)

    def _lock_loop(self):
        try:
            while True:
                try:
                    if (
                        self._previous_update_timestamp is not None
                        and time.time() - self._previous_update_timestamp
                        >= hub.constants.DATASET_LOCK_VALIDITY
                    ):
                        # Its been too long since last update, another machine might have locked the storage
                        lock_bytes = self.storage.get(self.path)
                        if lock_bytes:
                            nodeid, _, _ = _parse_lock_bytes(lock_bytes)
                            if nodeid != uuid.getnode():
                                if self.lock_lost_callback:
                                    self.lock_lost_callback()
                                self.acquired = False
                                return
                    self._previous_update_timestamp = time.time()
                    self.storage[self.path] = _get_lock_bytes(self.username)
                except Exception:
                    pass
                time.sleep(hub.constants.DATASET_LOCK_UPDATE_INTERVAL)
        except Exception:  # Thread termination
            return

    def acquire(self):
        if self.acquired:
            return
        self.storage.check_readonly()
        lock_bytes = self.storage.get(self.path)
        if lock_bytes is not None:
            nodeid = None
            try:
                nodeid, timestamp, _ = _parse_lock_bytes(lock_bytes)
            except Exception:  # parse error from corrupt lock file, ignore
                pass
            if nodeid:
                if nodeid == uuid.getnode():
                    # Lock left by this machine from a previous run, ignore
                    pass
                elif time.time() - timestamp < hub.constants.DATASET_LOCK_VALIDITY:
                    raise LockedException()

        self._thread = threading.Thread(target=self._lock_loop, daemon=True)
        self._thread.start()
        self.acquired = True

    def release(self):
        if not self.acquired:
            return
        with self._thread_lock:
            terminate_thread(self._thread)
            self._acquired = False
        try:
            del self.storage[self.path]
        except Exception:
            pass


_LOCKS: Dict[str, Lock] = {}
_REFS: Dict[str, Set[int]] = defaultdict(set)


def _get_lock_key(storage_path: str, commit_id: str):
    return storage_path + ":" + commit_id


def _get_lock_file_path(version: Optional[str] = None) -> str:
    if version in (None, FIRST_COMMIT_ID):
        return get_dataset_lock_key()
    return "versions/" + version + "/" + get_dataset_lock_key()  # type: ignore


def lock_dataset(
    dataset,
    lock_lost_callback: Optional[Callable] = None,
):
    """Locks a StorageProvider instance to avoid concurrent writes from multiple machines.

    Args:
        dataset: Dataset instance.
        lock_lost_callback (Callable, Optional): Called if the lock is lost after acquiring.

    Raises:
        LockedException: If the storage is already locked by a different machine.
    """
    storage = get_base_storage(dataset.storage)
    version = dataset.version_state["commit_id"]
    key = _get_lock_key(get_path_from_storage(storage), version)
    lock = _LOCKS.get(key)
    if lock:
        lock.acquire()
    else:
        lock = PersistentLock(
            storage,
            path=_get_lock_file_path(version),
            lock_lost_callback=lock_lost_callback,
        )
        _LOCKS[key] = lock
    _REFS[key].add(id(dataset))


def unlock_dataset(dataset):
    """Unlocks a storage provider that was locked by this machine.

    Args:
        dataset: The dataset to be unlocked
    """
    storage = get_base_storage(dataset.storage)
    version = dataset.version_state["commit_id"]
    key = _get_lock_key(get_path_from_storage(storage), version)
    try:
        lock = _LOCKS[key]
        ref_set = _REFS[key]
        try:
            ref_set.remove(id(dataset))
        except KeyError:
            pass
        if not ref_set:
            lock.release()
            del _REFS[key]
            del _LOCKS[key]
    except KeyError:
        pass



import hub
from math import ceil
import time
import boto3
import botocore  # type: ignore
import posixpath
from typing import Dict, Optional, Tuple
from botocore.session import ComponentLocator
from hub.client.client import HubBackendClient
from hub.core.storage.provider import StorageProvider
from hub.util.exceptions import (
    S3DeletionError,
    S3GetError,
    S3SetError,
    S3Error,
    PathNotEmptyException,
)
from hub.util.warnings import always_warn
from botocore.exceptions import (
    ReadTimeoutError,
    ConnectionError,
    ConnectTimeoutError,
    ConnectionClosedError,
    EndpointConnectionError,
    IncompleteReadError,
)

CONNECTION_ERRORS = (
    ReadTimeoutError,
    ConnectionError,
    ConnectTimeoutError,
    ConnectionClosedError,
    EndpointConnectionError,
    IncompleteReadError,
)

try:
    from botocore.exceptions import ResponseStreamingError

    CONNECTION_ERRORS = CONNECTION_ERRORS + (ResponseStreamingError,)  # type: ignore
except ImportError:
    pass


class S3ReloadCredentialsManager:
    """Tries to reload the credentials if the error is due to expired token, if error still occurs, it raises it."""

    def __init__(self, s3p, error_class: S3Error):
        self.error_class = error_class
        self.s3p = s3p

    def __enter__(self):
        self.s3p._locate_and_load_creds()
        self.s3p._set_s3_client_and_resource()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            raise self.error_class(exc_value).with_traceback(exc_traceback)


class S3ResetClientManager:
    """Resets the client, if error still occurs, it raises it."""

    def __init__(self, s3p, error_class: S3Error):
        self.error_class = error_class
        self.s3p = s3p

    def __enter__(self):
        self.s3p._check_update_creds(force=True)
        self.s3p._initialize_s3_parameters()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            raise self.error_class(exc_value).with_traceback(exc_traceback)


class S3Provider(StorageProvider):
    """Provider class for using S3 storage."""

    def __init__(
        self,
        root: str,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        aws_region: Optional[str] = None,
        profile_name: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """Initializes the S3Provider

        Example:

            >>> s3_provider = S3Provider("snark-test/benchmarks")

        Args:
            root (str): The root of the provider. All read/write request keys will be appended to root.
            aws_access_key_id (str, optional): Specifies the AWS access key used as part of the credentials to
                authenticate the user.
            aws_secret_access_key (str, optional): Specifies the AWS secret key used as part of the credentials to
                authenticate the user.
            aws_session_token (str, optional): Specifies an AWS session token used as part of the credentials to
                authenticate the user.
            endpoint_url (str, optional): The complete URL to use for the constructed client.
                This needs to be provided for cases in which you're interacting with MinIO, Wasabi, etc.
            aws_region (str, optional): Specifies the AWS Region to send requests to.
            profile_name (str, optional): Specifies the AWS profile name to use.
            token (str, optional): Activeloop token, used for fetching credentials for Hub datasets (if this is underlying storage for Hub dataset).
                This is optional, tokens are normally autogenerated.
        """
        self.root = root
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.aws_region: Optional[str] = aws_region
        self.endpoint_url: Optional[str] = endpoint_url
        self.expiration: Optional[str] = None
        self.tag: Optional[str] = None
        self.token: Optional[str] = token
        self.loaded_creds_from_environment = False
        self.client_config = hub.config["s3"]
        self.start_time = time.time()
        self.profile_name = profile_name
        self._initialize_s3_parameters()
        self._presigned_urls: Dict[str, Tuple[str, float]] = {}

    def subdir(self, path: str):
        sd = self.__class__(
            root=posixpath.join(self.root, path),
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            aws_region=self.aws_region,
            endpoint_url=self.endpoint_url,
        )
        if sd.expiration:
            sd._set_hub_creds_info(self.hub_path, self.expiration)  # type: ignore
        return sd

    def _set(self, path, content):
        self.client.put_object(
            Bucket=self.bucket,
            Body=content,
            Key=path,
            ContentType="application/octet-stream",  # signifies binary data
        )

    def __setitem__(self, path, content):
        """Sets the object present at the path with the value

        Args:
            path (str): the path relative to the root of the S3Provider.
            content (bytes): the value to be assigned at the path.

        Raises:
            S3SetError: Any S3 error encountered while setting the value at the path.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        self._check_update_creds()
        path = "".join((self.path, path))
        content = bytearray(memoryview(content))
        try:
            self._set(path, content)
        except botocore.exceptions.ClientError as err:
            reload = self.need_to_reload_creds(err)
            manager = S3ReloadCredentialsManager if reload else S3ResetClientManager
            with manager(self, S3SetError):
                self._set(path, content)
        except CONNECTION_ERRORS as err:
            tries = self.num_tries
            for i in range(1, tries + 1):
                always_warn(f"Encountered connection error, retry {i} out of {tries}")
                try:
                    self._set(path, content)
                    always_warn(
                        f"Connection re-established after {i} {['retries', 'retry'][i==1]}."
                    )
                    return
                except Exception:
                    pass
            raise S3SetError(err) from err
        except Exception as err:
            raise S3SetError(err) from err

    def _get(self, path, bucket=None):
        bucket = bucket or self.bucket
        resp = self.client.get_object(
            Bucket=bucket,
            Key=path,
        )
        return resp["Body"].read()

    def __getitem__(self, path):
        """Gets the object present at the path.

        Args:
            path (str): the path relative to the root of the S3Provider.

        Returns:
            bytes: The bytes of the object present at the path.

        Raises:
            KeyError: If an object is not found at the path.
            S3GetError: Any other error other than KeyError while retrieving the object.
        """
        return self.get_bytes(path)

    def _get_bytes(self, path, start_byte: int = None, end_byte: int = None):
        if start_byte is not None and end_byte is not None:
            if start_byte == end_byte:
                return b""
            range = f"bytes={start_byte}-{end_byte - 1}"
        elif start_byte is not None:
            range = f"bytes={start_byte}-"
        elif end_byte is not None:
            range = f"bytes=0-{end_byte - 1}"
        else:
            range = ""
        resp = self.client.get_object(Bucket=self.bucket, Key=path, Range=range)
        return resp["Body"].read()

    def get_bytes(
        self,
        path: str,
        start_byte: Optional[int] = None,
        end_byte: Optional[int] = None,
    ):
        """Gets the object present at the path within the given byte range.

        Args:
            path (str): The path relative to the root of the provider.
            start_byte (int, optional): If only specific bytes starting from ``start_byte`` are required.
            end_byte (int, optional): If only specific bytes up to end_byte are required.

        Returns:
            bytes: The bytes of the object present at the path within the given byte range.

        Raises:
            InvalidBytesRequestedError: If ``start_byte`` > ``end_byte`` or ``start_byte`` < 0 or ``end_byte`` < 0.
            KeyError: If an object is not found at the path.
            S3GetError: Any other error other than KeyError while retrieving the object.
        """
        self._check_update_creds()
        path = "".join((self.path, path))
        try:
            return self._get_bytes(path, start_byte, end_byte)
        except botocore.exceptions.ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchKey":
                raise KeyError(err) from err
            reload = self.need_to_reload_creds(err)
            manager = S3ReloadCredentialsManager if reload else S3ResetClientManager  # type: ignore
            with manager(self, S3GetError):  # type: ignore
                return self._get_bytes(path, start_byte, end_byte)
        except CONNECTION_ERRORS as err:
            tries = self.num_tries
            for i in range(1, tries + 1):
                always_warn(f"Encountered connection error, retry {i} out of {tries}")
                try:
                    ret = self._get_bytes(path, start_byte, end_byte)
                    always_warn(
                        f"Connection re-established after {i} {['retries', 'retry'][i==1]}."
                    )
                    return ret
                except Exception:
                    pass
            raise S3GetError(err) from err
        except Exception as err:
            raise S3GetError(err) from err

    def _del(self, path):
        self.client.delete_object(Bucket=self.bucket, Key=path)

    def __delitem__(self, path):
        """Delete the object present at the path.

        Args:
            path (str): the path to the object relative to the root of the S3Provider.

        Note:
            If the object is not found, s3 won't raise KeyError.

        Raises:
            S3DeletionError: Any S3 error encountered while deleting the object.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        self._check_update_creds()
        path = "".join((self.path, path))
        try:
            self._del(path)
        except botocore.exceptions.ClientError as err:
            reload = self.need_to_reload_creds(err)
            manager = S3ReloadCredentialsManager if reload else S3ResetClientManager
            with manager(self, S3DeletionError):
                self._del(path)
        except CONNECTION_ERRORS as err:
            tries = self.num_tries
            for i in range(1, tries + 1):
                always_warn(f"Encountered connection error, retry {i} out of {tries}")
                try:
                    self._del(path)
                    always_warn(
                        f"Connection re-established after {i} {['retries', 'retry'][i==1]}."
                    )
                    return
                except Exception:
                    pass
            raise S3DeletionError(err) from err
        except Exception as err:
            raise S3DeletionError(err) from err

    @property
    def num_tries(self):
        return min(ceil((time.time() - self.start_time) / 300), 5)

    def _keys_iterator(self):
        self._check_update_creds()
        prefix = self.path
        start_after = ""
        prefix = prefix[1:] if prefix.startswith("/") else prefix
        start_after = (start_after or prefix) if prefix.endswith("/") else start_after
        paginator = self.client.get_paginator("list_objects_v2")
        for page in paginator.paginate(
            Bucket=self.bucket, Prefix=prefix, StartAfter=start_after
        ):
            for content in page.get("Contents", ()):
                yield content["Key"]

    def _all_keys(self):
        """Helper function that lists all the objects present at the root of the S3Provider.

        Returns:
            set: set of all the objects found at the root of the S3Provider.

        Raises:
            S3ListError: Any S3 error encountered while listing the objects.
        """
        len_path = len(self.path.split("/")) - 1
        return ("/".join(name.split("/")[len_path:]) for name in self._keys_iterator())

    def __len__(self):
        """Returns the number of files present at the root of the S3Provider.

        Note:
            This is an expensive operation.

        Returns:
            int: the number of files present inside the root.

        Raises:
            S3ListError: Any S3 error encountered while listing the objects.
        """
        return sum(1 for _ in self._keys_iterator())

    def __iter__(self):
        """Generator function that iterates over the keys of the S3Provider.

        Yields:
            str: the name of the object that it is iterating over.
        """
        self._check_update_creds()
        yield from self._all_keys()

    def clear(self, prefix=""):
        """Deletes ALL data with keys having given prefix on the s3 bucket (under self.root).

        Warning:
            Exercise caution!
        """
        self.check_readonly()
        self._check_update_creds()
        path = posixpath.join(self.path, prefix) if prefix else self.path
        if self.resource is not None:
            try:
                bucket = self.resource.Bucket(self.bucket)
                bucket.objects.filter(Prefix=path).delete()
            except Exception as err:
                reload = self.need_to_reload_creds(err)
                manager = S3ReloadCredentialsManager if reload else S3ResetClientManager
                with manager(self, S3DeletionError):
                    bucket = self.resource.Bucket(self.bucket)
                    bucket.objects.filter(Prefix=self.path).delete()

        else:
            super().clear()

    def rename(self, root):
        """Rename root folder."""
        self.check_readonly()
        self._check_update_creds()
        items = []
        paginator = self.client.get_paginator("list_objects_v2")
        pages = paginator.paginate(Bucket=self.bucket, Prefix=self.path)
        for page in pages:
            items.extend(page["Contents"])
        path = root.replace("s3://", "")
        _, new_path = path.split("/", 1)
        try:
            dest_objects = self.client.list_objects_v2(
                Bucket=self.bucket, Prefix=new_path
            )["Contents"]
            for item in dest_objects:
                raise PathNotEmptyException(use_hub=False)
        except KeyError:
            pass
        for item in items:
            old_key = item["Key"]
            copy_source = {"Bucket": self.bucket, "Key": old_key}
            new_key = "/".join([new_path, posixpath.relpath(old_key, self.path)])
            self.client.copy_object(
                CopySource=copy_source, Bucket=self.bucket, Key=new_key
            )
            self.client.delete_object(Bucket=self.bucket, Key=old_key)

        self.root = root
        self.path = new_path
        if not self.path.endswith("/"):
            self.path += "/"

    def _state_keys(self):
        """Keys used to store the state of the provider."""
        return {
            "root",
            "aws_access_key_id",
            "aws_secret_access_key",
            "aws_session_token",
            "aws_region",
            "endpoint_url",
            "client_config",
            "expiration",
            "tag",
            "token",
            "loaded_creds_from_environment",
            "read_only",
            "profile_name",
        }

    def __getstate__(self):
        return {key: getattr(self, key) for key in self._state_keys()}

    def __setstate__(self, state):
        assert set(state.keys()) == self._state_keys()
        self.__dict__.update(state)
        self.start_time = time.time()
        self._initialize_s3_parameters()

    def _set_bucket_and_path(self):
        root = self.root.replace("s3://", "")
        split_root = root.split("/", 1)
        self.bucket = split_root[0]
        if len(split_root) > 1:
            self.path = split_root[1]
        else:
            self.path = ""
        if not self.path.endswith("/"):
            self.path += "/"

    def _set_hub_creds_info(self, hub_path: str, expiration: str):
        """Sets the tag and expiration of the credentials. These are only relevant to datasets using Hub storage.
        This info is used to fetch new credentials when the temporary 12 hour credentials expire.

        Args:
            hub_path (str): The hub cloud path to the dataset.
            expiration (str): The time at which the credentials expire.
        """
        self.hub_path = hub_path
        self.tag = hub_path[6:]  # removing the hub:// part from the path
        self.expiration = expiration

    def _initialize_s3_parameters(self):
        self._set_bucket_and_path()

        if self.aws_access_key_id is None and self.aws_secret_access_key is None:
            self._locate_and_load_creds()
            self.loaded_creds_from_environment = True

        self._set_s3_client_and_resource()

    def _check_update_creds(self, force=False):
        """If the client has an expiration time, check if creds are expired and fetch new ones.
        This would only happen for datasets stored on Hub storage for which temporary 12 hour credentials are generated.
        """
        if self.expiration and (force or float(self.expiration) < time.time()):
            client = HubBackendClient(self.token)
            org_id, ds_name = self.tag.split("/")

            mode = "r" if self.read_only else "a"

            url, creds, mode, expiration = client.get_dataset_credentials(
                org_id,
                ds_name,
                mode,
                force,
            )
            self.expiration = expiration
            self._set_s3_client_and_resource(
                creds.get("aws_access_key_id"),
                creds.get("aws_secret_access_key"),
                creds.get("aws_session_token"),
            )

    def _locate_and_load_creds(self):
        session = boto3.session.Session(profile_name=self.profile_name)._session
        component_locator = ComponentLocator()
        component_locator.lazy_register_component(
            "credential_provider", session._create_credential_resolver
        )
        credentials = component_locator.get_component(
            "credential_provider"
        ).load_credentials()
        if credentials is not None:
            self.aws_access_key_id = credentials.access_key
            self.aws_secret_access_key = credentials.secret_key
            self.aws_session_token = credentials.token
            self.aws_region = session._resolve_region_name(
                self.aws_region, self.client_config
            )

    def _set_s3_client_and_resource(
        self, aws_access_key_id=None, aws_secret_access_key=None, aws_session_token=None
    ):
        key = aws_access_key_id or self.aws_access_key_id
        secret = aws_secret_access_key or self.aws_secret_access_key
        token = aws_session_token or self.aws_session_token
        args = {
            "aws_access_key_id": key,
            "aws_secret_access_key": secret,
            "aws_session_token": token,
            "region_name": self.aws_region,
            "endpoint_url": self.endpoint_url,
            "config": self.client_config,
        }
        session = boto3.session.Session(profile_name=self.profile_name)
        self.client = session.client("s3", **args)
        self.resource = session.resource("s3", **args)

    def need_to_reload_creds(self, err: botocore.exceptions.ClientError) -> bool:
        """Checks if the credentials need to be reloaded.
        This happens if the credentials were loaded from the environment and have now expired.
        """
        return (
            err.response["Error"]["Code"] == "ExpiredToken"
            and self.loaded_creds_from_environment
        )

    def get_presigned_url(self, key, full=False):
        self._check_update_creds()
        if full:
            root = key.replace("s3://", "")
            split_root = root.split("/", 1)
            bucket = split_root[0]
            path = split_root[1] if len(split_root) > 1 else ""
        else:
            bucket = self.bucket
            path = "".join((self.path, key))

        url = None
        cached = self._presigned_urls.get(path)
        if cached:
            url, t_store = cached
            t_now = time.time()
            if t_now - t_store > 3200:
                del self._presigned_urls[path]
                url = None

        if url is None:
            if self._is_hub_path:
                client = HubBackendClient(self.token)
                org_id, ds_name = self.tag.split("/")  # type: ignore
                url = client.get_presigned_url(org_id, ds_name, key)
            else:
                url = self.client.generate_presigned_url(
                    "get_object",
                    Params={"Bucket": bucket, "Key": path},
                    ExpiresIn=3600,
                )
            self._presigned_urls[path] = (url, time.time())
        return url

    def get_object_size(self, path: str) -> int:
        path = "".join((self.path, path))
        obj = self.resource.Object(self.bucket, path)
        return obj.content_length

    def get_object_from_full_url(self, url: str):
        root = url.replace("s3://", "")
        split_root = root.split("/", 1)
        bucket = split_root[0]
        path = split_root[1] if len(split_root) > 1 else ""
        try:
            return self._get(path, bucket)
        except botocore.exceptions.ClientError as err:
            if err.response["Error"]["Code"] == "NoSuchKey":
                raise KeyError(err) from err
            reload = self.need_to_reload_creds(err)
            manager = S3ReloadCredentialsManager if reload else S3ResetClientManager  # type: ignore
            with manager(self, S3GetError):  # type: ignore
                return self._get(path, bucket)
        except CONNECTION_ERRORS as err:
            tries = self.num_tries
            for i in range(1, tries + 1):
                always_warn(f"Encountered connection error, retry {i} out of {tries}")
                try:
                    ret = self._get(path, bucket)
                    always_warn(
                        f"Connection re-established after {i} {['retries', 'retry'][i==1]}."
                    )
                    return ret
                except Exception:
                    pass
            raise S3GetError(err) from err
        except Exception as err:
            raise S3GetError(err) from err



import sys
import hub
import pytest
from io import StringIO
from contextlib import contextmanager
from click.testing import CliRunner
from hub.cli.auth import login, logout
from hub.client.client import HubBackendClient
from hub.util.exceptions import AgreementNotAcceptedError, NotLoggedInAgreementError


@contextmanager
def replace_stdin(target):
    orig = sys.stdin
    sys.stdin = target
    yield
    sys.stdin = orig


def dont_agree(path):
    """Load the hub cloud dataset at path and simulate disagreeing to the terms of access."""

    with pytest.raises(AgreementNotAcceptedError):
        # this text can be anything except expected
        with replace_stdin(StringIO("no, i don't agree!")):
            hub.load(path)


def agree(path):
    """Load the hub cloud dataset at path and simulate agreeing to the terms of access."""
    dataset_name = path.split("/")[-1]
    with replace_stdin(StringIO(dataset_name)):
        ds = hub.load(path)
    ds.images[0].numpy()

    # next load should work without agreeing
    ds = hub.load(path)
    ds.images[0].numpy()


def reject(path):
    client = HubBackendClient()
    org_id, ds_name = path.split("/")[-2:]
    client.reject_agreements(org_id, ds_name)


def test_agreement_logged_out(hub_cloud_dev_credentials):
    runner = CliRunner()
    runner.invoke(logout)
    path = "hub://activeloop/imagenet-test"
    with pytest.raises(NotLoggedInAgreementError):
        agree(path)


def test_agreement_logged_in(hub_cloud_dev_credentials):
    runner = CliRunner()
    username, password = hub_cloud_dev_credentials
    runner.invoke(login, f"-u {username} -p {password}")
    path = "hub://activeloop/imagenet-test"
    agree(path)
    reject(path)
    runner.invoke(logout)


def test_not_agreement_logged_in(hub_cloud_dev_credentials):
    runner = CliRunner()
    username, password = hub_cloud_dev_credentials
    runner.invoke(login, f"-u {username} -p {password}")
    path = "hub://activeloop/imagenet-test"
    dont_agree(path)
    runner.invoke(logout)



from hub.core.meta.encode.base_encoder import Encoder, LAST_SEEN_INDEX_COLUMN
from hub.constants import ENCODING_DTYPE
from typing import Tuple
from hub.core.storage.provider import StorageProvider
import numpy as np


class ShapeEncoder(Encoder):
    def _derive_value(self, row: np.ndarray, *_) -> Tuple:  # type: ignore
        return tuple(row[:LAST_SEEN_INDEX_COLUMN])

    @property
    def dimensionality(self) -> int:
        return len(self[0])

    def _validate_incoming_item(self, shape: Tuple[int], _):
        if self.num_samples > 0:
            last_shape = self[-1]  # TODO: optimize this

            if len(shape) != len(last_shape):
                raise ValueError(
                    f"All sample shapes in a tensor must have the same len(shape). Expected: {len(last_shape)} got: {len(shape)}."
                )

        super()._validate_incoming_item(shape, _)

    def _combine_condition(
        self, shape: Tuple[int], compare_row_index: int = -1
    ) -> bool:
        last_shape = self._derive_value(self._encoded[compare_row_index])
        return shape == last_shape



from typing import Any, Dict, List, Optional, Set, Tuple

from hub.core.storage import LRUCache
from hub.core.version_control.commit_diff import CommitDiff
from hub.core.version_control.commit_node import CommitNode
from hub.util.diff import sanitize_commit
from hub.util.exceptions import TensorModifiedError
from hub.util.keys import get_tensor_commit_diff_key


def get_modified_indexes(
    tensor: str,
    current_commit_id: str,
    target_id: Optional[str],
    version_state: Dict[str, Any],
    storage: LRUCache,
) -> List[int]:
    if target_id is None:
        indexes, _ = get_modified_indexes_for_commit(tensor, current_commit_id, storage)
        return sorted(list(indexes))

    indexes = set()
    target_id = sanitize_commit(target_id, version_state)
    commit_node_map = version_state["commit_node_map"]
    current_node: CommitNode = commit_node_map[current_commit_id]
    target_node: CommitNode = commit_node_map[target_id]

    if not check_ancestor(current_node, target_node):
        raise TensorModifiedError

    while current_node.commit_id != target_node.commit_id:
        commit_id = current_node.commit_id
        idxes, stop = get_modified_indexes_for_commit(tensor, commit_id, storage)
        indexes.update(idxes)
        if stop:
            break
        current_node = current_node.parent  # type: ignore
    return sorted(list(indexes))


def get_modified_indexes_for_commit(
    tensor: str, commit_id: str, storage: LRUCache
) -> Tuple[Set[int], bool]:
    indexes: Set[int] = set()
    try:
        commit_diff_key = get_tensor_commit_diff_key(tensor, commit_id)
        commit_diff: CommitDiff = storage.get_hub_object(commit_diff_key, CommitDiff)

        data_added = range(*commit_diff.data_added)
        data_updated = commit_diff.data_updated

        indexes.update(data_added)
        indexes.update(data_updated)

        stop = commit_diff.data_transformed
        return indexes, stop
    except KeyError:
        return indexes, False


def check_ancestor(current_node: CommitNode, target_node: CommitNode) -> bool:
    """Checks if the target node is an ancestor of the current node."""
    target_id = target_node.commit_id
    while current_node is not None:
        if current_node.commit_id == target_id:
            return True
        current_node = current_node.parent  # type: ignore
    return False



from typing import Callable, Dict, Optional
from uuid import uuid4

CREATE_DATASET_HOOKS: Dict[str, Callable] = {}
LOAD_DATASET_HOOKS: Dict[str, Callable] = {}
WRITE_DATASET_HOOKS: Dict[str, Callable] = {}
READ_DATASET_HOOKS: Dict[str, Callable] = {}
COMMIT_DATASET_HOOKS: Dict[str, Callable] = {}
HOOK_EVENTS: Dict[str, str] = {}


def _add_hook(event: str, hook: Callable, id: Optional[str] = None):
    if id is None:
        id = str(uuid4())
    if id in HOOK_EVENTS:
        raise Exception(f"Hook with id {id} already exists.")
    globals()[f"{event}_HOOKS"][id] = hook
    HOOK_EVENTS[id] = event


def add_create_dataset_hook(hook: Callable, id: Optional[str] = None):
    _add_hook("CREATE_DATASET", hook, id)


def add_load_dataset_hook(hook: Callable, id: Optional[str] = None):
    _add_hook("LOAD_DATASET", hook, id)


def add_read_dataset_hook(hook: Callable, id: Optional[str] = None):
    _add_hook("READ_DATASET", hook, id)


def add_write_dataset_hook(hook: Callable, id: Optional[str] = None):
    _add_hook("WRITE_DATASET", hook, id)


def add_commit_dataset_hook(hook: Callable, id: Optional[str] = None):
    _add_hook("COMMIT_DATASET", hook, id)


def remove_hook(id: str):
    del globals()[f"_{HOOK_EVENTS.pop(id)}_HOOKS"][id]


def dataset_created(ds):
    [f(ds) for f in CREATE_DATASET_HOOKS.values()]


def dataset_loaded(ds):
    [f(ds) for f in LOAD_DATASET_HOOKS.values()]


def dataset_written(ds):
    [f(ds) for f in WRITE_DATASET_HOOKS.values()]


def dataset_read(ds):
    [f(ds) for f in READ_DATASET_HOOKS.values()]


def dataset_committed(ds):
    [f(ds) for f in COMMIT_DATASET_HOOKS.values()]



import time
import os
from hub.core.storage.provider import StorageProvider
from io import BytesIO
import posixpath
import pickle
from typing import Dict, Optional, Union
from hub.util.hash import hash_inputs
import logging

try:
    from httplib2 import Http  # type: ignore
    from googleapiclient import discovery  # type: ignore
    from googleapiclient.http import (  # type: ignore
        MediaIoBaseDownload,
        MediaIoBaseUpload,
    )
    from googleapiclient.errors import HttpError  # type: ignore
    from google.auth.transport.requests import Request  # type: ignore
    from google_auth_oauthlib.flow import InstalledAppFlow  # type: ignore
    from google.oauth2.credentials import Credentials  # type: ignore

    _GDRIVE_PACKAGES_INSTALLED = True
except ImportError:
    _GDRIVE_PACKAGES_INSTALLED = False


logging.getLogger("googleapiclient.discovery_cache").setLevel(logging.ERROR)

SCOPES = [
    "https://www.googleapis.com/auth/drive.file",
    "https://www.googleapis.com/auth/drive.install",
    "https://www.googleapis.com/auth/drive.readonly",
]
FOLDER = "application/vnd.google-apps.folder"
FILE = "application/octet-stream"


class GoogleDriveIDManager:
    """Class used to make google drive path to id maps"""

    def __init__(self, drive, root: str, root_id: Optional[str] = None):
        self.path_id_map: Dict[str, str] = {}
        self.drive = drive
        self.root_path = root
        self.root_id = root_id if root_id else self.find_id(root)

    def find_id(self, path):
        """Find google drive id given path of folder"""

        dirname, basename = posixpath.split(path)
        try:
            file_list = (
                self.drive.files()
                .list(
                    q=f"""'{self.find_id(dirname) if dirname else 'root'}' in parents and 
                        name = '{basename}' and 
                        trashed = false""",
                    supportsAllDrives="true",
                    includeItemsFromAllDrives="true",
                    spaces="drive",
                    fields="files(id)",
                )
                .execute()
            )
        except HttpError:
            file_list = {"files": []}

        if len(file_list["files"]) > 0:
            id = file_list["files"][0].get("id")
        else:
            id = None
        return id

    def makemap(self, root_id, root_path):
        """Make mapping from google drive paths to ids for all files and folders under root"""

        file_list = (
            self.drive.files()
            .list(
                q=f"'{root_id}' in parents and trashed = false",
                spaces="drive",
                fields="files(name, id)",
            )
            .execute()["files"]
        )
        prefix = "" if root_path == self.root_path else root_path
        for file in file_list:
            path = posixpath.join(prefix, file["name"])
            self.path_id_map[path] = file["id"]
            self.makemap(file["id"], path)
        return self.path_id_map


class GDriveProvider(StorageProvider):
    """Provider class for using Google Drive storage."""

    client_id = None
    client_secret = None
    refresh_token = None

    def __init__(
        self, root: str, token: Optional[Union[str, Dict]] = None, makemap: bool = True
    ):
        """Initializes the GDriveProvider

        Example:

            >>> gdrive_provider = GDriveProvider("gdrive://folder_name/folder_name")

        Args:
            root(str): The root of the provider. All read/write request keys will be appended to root.
            token(dict, str, optional): Google Drive token. Can be path to the token file or the actual credentials dictionary.
            makemap(bool): Creates path to id map if ``True``.

        Note:
            - Requires ``client_secrets.json`` in working directory if ``token`` is not provided.
            - Due to limits on requests per 100 seconds on google drive api, continuous requests such as uploading many small files can be slow.
            - Users can request to increse their quotas on their google cloud platform.
        """

        creds = None

        if not token:
            token = "gdrive_token.json"

        if isinstance(token, str):
            if os.path.exists(token):
                creds = Credentials.from_authorized_user_file(token, SCOPES)
        elif isinstance(token, dict):
            creds = Credentials.from_authorized_user_info(token, SCOPES)

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                if os.path.exists("client_secrets.json"):
                    flow = InstalledAppFlow.from_client_secrets_file(
                        "client_secrets.json", SCOPES
                    )
                else:
                    OAUTH_CLIENT_ID = os.getenv("CLIENT_ID")
                    OAUTH_CLIENT_SECRET = os.getenv("CLIENT_SECRET")
                    flow = InstalledAppFlow.from_client_config(
                        {
                            "installed": {
                                "client_id": OAUTH_CLIENT_ID,
                                "client_secret": OAUTH_CLIENT_SECRET,
                                "redirect_uris": [
                                    "urn:ietf:wg:oauth:2.0:oob",
                                    "http://localhost",
                                ],
                                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                                "token_uri": "https://oauth2.googleapis.com/token",
                                "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                            }
                        },
                        scopes=SCOPES,
                    )
                creds = flow.run_local_server(port=0)

            with open("gdrive_token.json", "w") as token_file:
                token_file.write(creds.to_json())

        self.root = root
        if creds:
            self.client_id = creds.client_id
            self.client_secret = creds.client_secret
            self.refresh_token = creds.refresh_token

        self._init_provider(creds, makemap=makemap)

    def _init_provider(self, creds, makemap=True):
        self.drive = discovery.build("drive", "v3", credentials=creds)
        self.root_path = self.root.replace("gdrive://", "")
        if self.root_path == "":
            self.root_id = "root"
        if hasattr(self, "root_id"):
            self.gid = GoogleDriveIDManager(self.drive, self.root_path, self.root_id)  # type: ignore
        else:
            self.gid = GoogleDriveIDManager(self.drive, self.root_path)
            self.root_id = self.gid.root_id
        if self.root_id is None:
            self.root_id = "root"
            root_dir = self.make_dir(self.root_path, find=True)
            self.root_id = self.gid.root_id = root_dir.get("id")
            for i in range(len(self.root_path.split("/"))):
                self.gid.path_id_map.pop(
                    self.root_path.split("/", i)[0], None
                )  # Remove root dir components from map
        if makemap:
            self.gid.makemap(self.root_id, self.root_path)

    def _init_from_state(self):
        token = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": self.refresh_token,
        }
        creds = Credentials.from_authorized_user_info(token, SCOPES)
        self._init_provider(creds)

    def _get_id(self, path):
        return self.gid.path_id_map.get(path)

    def _pop_id(self, path):
        return self.gid.path_id_map.pop(path)

    def _set_id(self, path, id):
        self.gid.path_id_map[path] = id

    def make_dir(self, path, find=False):
        dirname, basename = posixpath.split(path)
        if dirname:
            if find:
                parent_id = self.gid.find_id(dirname)
            else:
                parent_id = self._get_id(dirname)
            if not parent_id:
                locked = self._lock_creation(dirname)
                if locked:
                    self.make_dir(dirname)
                    self._unlock_creation(dirname)
                parent_id = self._get_id(dirname)
            folder = self._create_file(basename, FOLDER, parent_id)
        else:
            folder = self._create_file(basename, FOLDER, self.root_id)

        self._set_id(path, folder.get("id"))
        return folder

    def _create_file(self, name, mimeType, parent=None, content=None):
        file_metadata = {
            "name": name,
            "mimeType": mimeType,
            "parents": [parent if parent else self.root_id],
        }

        if not content:
            file = self.drive.files().create(body=file_metadata, fields="id").execute()
        else:
            content = MediaIoBaseUpload(BytesIO(content), mimeType)
            file = (
                self.drive.files()
                .create(body=file_metadata, media_body=content, fields="id")
                .execute()
            )

        return file

    def _write_to_file(self, id, content):
        content = MediaIoBaseUpload(BytesIO(content), FILE)
        file = self.drive.files().update(media_body=content, fileId=id).execute()
        return file

    def _delete_file(self, id):
        file = self.drive.files().delete(fileId=id).execute()
        return file

    def sync(self):
        """Sync provider keys with actual storage"""
        self.gid.makemap(self.root_id, self.root_path)

    def get_object_by_id(self, id):
        request = self.drive.files().get_media(fileId=id)
        file = BytesIO()
        downloader = MediaIoBaseDownload(file, request)
        done = False
        while not done:
            status, done = downloader.next_chunk()

        file.seek(0)
        return file.read()

    def __getitem__(self, path):
        id = self._get_id(path)
        if not id:
            raise KeyError(path)

        return self.get_object_by_id(id)

    def get_object_from_full_url(self, url: str):
        url = url.replace("gdrive://", "")
        id = self.gid.find_id(url)
        return self.get_object_by_id(id)

    def _lock_creation(self, path):
        # lock creation of folder, otherwise multiple workers can create folders of the same name.
        lock_hash = "." + hash_inputs(self.root_id, path)
        try:
            lock_file = open(lock_hash, "x")
            lock_file.close()
            return True
        except FileExistsError:
            while os.path.exists(lock_hash):
                time.sleep(0.1)
            return False

    def _unlock_creation(self, path):
        lock_hash = "." + hash_inputs(self.root_id, path)
        os.remove(lock_hash)

    def __setitem__(self, path, content):
        self.check_readonly()
        id = self._get_id(path)
        if not id:
            dirname, basename = posixpath.split(path)
            if dirname:
                parent_id = self._get_id(dirname)
                if not parent_id:
                    locked = self._lock_creation(dirname)
                    if locked:
                        self.make_dir(dirname)
                        self._unlock_creation(dirname)
                    self.sync()
                    parent_id = self._get_id(dirname)
            else:
                parent_id = self.root_id
            file = self._create_file(basename, FILE, parent_id, content)
            self._set_id(path, file.get("id"))
            return

        self._write_to_file(id, content)
        return

    def __delitem__(self, path):
        self.check_readonly()
        id = self._pop_id(path)
        if not id:
            raise KeyError(path)
        self._delete_file(id)

    def __getstate__(self):
        return (
            self.root,
            self.root_id,
            self.client_id,
            self.client_secret,
            self.refresh_token,
        )

    def __setstate__(self, state):
        self.root = state[0]
        self.root_id = state[1]
        self.client_id = state[2]
        self.client_secret = state[3]
        self.refresh_token = state[4]
        self._init_from_state()

    def _all_keys(self):
        keys = set(self.gid.path_id_map.keys())
        return keys

    def __iter__(self):
        yield from self._all_keys()

    def __len__(self):
        return len(self._all_keys())

    def clear(self, prefix=""):
        self.check_readonly()
        for key in self._all_keys():
            if key.startswith(prefix):
                try:
                    del self[key]
                except:
                    pass
        if not prefix:
            self._delete_file(self.root_id)



from typing import Dict, List

import numpy as np


POINT_CLOUD_FIELD_NAME_TO_TYPESTR: Dict = {
    "X": "<i4",
    "Y": "<i4",
    "Z": "<i4",
    "intensity": "<u2",
    "return_number": "|u1",
    "number_of_returns": "|u1",
    "scan_direction_flag": "|u1",
    "edge_of_flight_line": "|u1",
    "classification": "|u1",
    "synthetic": "|u1",
    "key_point": "|u1",
    "withheld": "|u1",
    "scan_angle_rank": "|i1",
    "user_data": "|u1",
    "point_source_id": "<u2",
    "red": "<u2",
    "green": "<u2",
    "blue": "<u2",
}


def default_version_parser(point_cloud):
    return {
        "major": point_cloud.header.DEFAULT_VERSION.major,
        "minor": point_cloud.header.DEFAULT_VERSION.minor,
    }


def creation_date_parser(point_cloud):
    return {
        "year": point_cloud.header.creation_date.year,
        "month": point_cloud.header.creation_date.month,
        "day": point_cloud.header.creation_date.day,
    }


def global_encoding_parser(point_cloud):
    return {
        "GPS_TIME_TYPE_MASK": point_cloud.header.global_encoding.GPS_TIME_TYPE_MASK,
        "SYNTHETIC_RETURN_NUMBERS_MASK": point_cloud.header.global_encoding.SYNTHETIC_RETURN_NUMBERS_MASK,
        "WAVEFORM_EXTERNAL_MASK": point_cloud.header.global_encoding.WAVEFORM_EXTERNAL_MASK,
        "WAVEFORM_INTERNAL_MASK": point_cloud.header.global_encoding.WAVEFORM_INTERNAL_MASK,
        "WKT_MASK": point_cloud.header.global_encoding.WKT_MASK,
        "gps_time_type": point_cloud.header.global_encoding.gps_time_type,
        "synthetic_return_numbers": point_cloud.header.global_encoding.synthetic_return_numbers,
        "value": point_cloud.header.global_encoding.value,
        "waveform_data_packets_external": point_cloud.header.global_encoding.waveform_data_packets_external,
        "waveform_data_packets_internal": point_cloud.header.global_encoding.waveform_data_packets_internal,
        "wkt": point_cloud.header.global_encoding.wkt,
    }


LAS_HEADER_FILED_NAME_TO_PARSER = {
    "DEFAULT_VERSION": default_version_parser,
    "version": default_version_parser,
    "creation_date": creation_date_parser,
    "global_encoding": global_encoding_parser,
}


def cast_point_cloud_array_to_proper_dtype(
    full_arr, sample_index, dimension_index, dtype
):
    if isinstance(full_arr, List):
        return full_arr[sample_index][:, dimension_index].astype(np.dtype(dtype))
    return full_arr[sample_index, :, dimension_index].astype(np.dtype(dtype))



import time


def expires_in_to_expires_at(creds: dict) -> None:
    if "expires_in" not in creds:
        return
    creds["expires_at"] = time.time() + creds.pop("expires_in")


def is_expired_token(creds: dict) -> bool:
    if "Authorization" not in creds:
        return False
    if "expires_at" not in creds:
        return False
    return creds["expires_at"] < time.time()



import os
from hub.util.exceptions import InvalidPathException
import numpy as np
from .base import StructuredDataset
from hub.core.dataset import Dataset
from tqdm import tqdm  # type: ignore


class DataFrame(StructuredDataset):
    def __init__(self, source):
        """Convert a pandas dataframe to a Hub dataset.

        Args:
            source: Pandas dataframe object.

        Raises:
            Exception: If source is not a pandas dataframe object.
        """
        import pandas as pd  # type: ignore

        super().__init__(source)
        if not isinstance(self.source, pd.DataFrame):
            raise Exception("Source is not a pandas dataframe object.")

    def fill_dataset(self, ds: Dataset, use_progress_bar: bool = True) -> Dataset:
        """Fill dataset with data from the dataframe - one tensor per column

        Args:
            ds (Dataset) : A Hub dataset object.
            use_progress_bar (bool) : Defines if the method uses a progress bar. Defaults to True.

        Returns:
            A hub dataset.

        """
        keys = list(self.source.columns)
        skipped_keys: list = []
        iterator = tqdm(
            keys,
            desc="Ingesting... (%i keys skipped)" % (len(skipped_keys)),
            disable=not use_progress_bar,
        )
        with ds, iterator:
            for key in iterator:
                try:
                    dtype = self.source[key].dtype
                    if dtype == np.dtype("object"):
                        self.source[key].fillna("", inplace=True)
                        ds.create_tensor(key, dtype=str, htype="text")
                    else:
                        self.source[key].fillna(0, inplace=True)
                        ds.create_tensor(key)
                    ds[key].extend(self.source[key].values.tolist())
                except Exception as e:
                    skipped_keys.append(key)
                    iterator.set_description(
                        "Ingesting... (%i keys skipped)" % (len(skipped_keys))
                    )
                    continue

        return ds



from hub.core.storage.hub_memory_object import HubMemoryObject
from hub.core.storage import LRUCache
from hub.util.keys import get_dataset_diff_key
import typing
from collections import OrderedDict


class DatasetDiff(HubMemoryObject):
    def __init__(self) -> None:
        self.is_dirty = False
        self.info_updated = False
        self.renamed: typing.OrderedDict = OrderedDict()
        self.deleted: typing.List[str] = []

    def tobytes(self) -> bytes:
        """Returns bytes representation of the dataset diff

        The format stores the following information in order:
        1. The first byte is a boolean value indicating whether the Dataset info was modified or not.
        2. The next 8 bytes give the number of renamed tensors, let's call this m.
        3. Next, there will be m blocks of bytes with the following format:
            1. 8 + 8 bytes giving the length of old and new names, let's call them x and y.
            2. x bytes of old name.
            3. y bytes of new name.
        4. The next 8 bytes give the number of deleted tensors, let's call this n.
        5. Next, there will be n blocks of bytes with the following format:
            1. 8 bytes giving the length of the name of the deleted tensor, let's call this z.
            2. n bytes of name of the deleted tensor.
        """
        return b"".join(
            [
                self.info_updated.to_bytes(1, "big"),
                len(self.renamed).to_bytes(8, "big"),
                *(
                    b"".join(
                        [
                            len(old_name).to_bytes(8, "big"),
                            len(new_name).to_bytes(8, "big"),
                            (old_name + new_name).encode("utf-8"),
                        ]
                    )
                    for old_name, new_name in self.renamed.items()
                ),
                len(self.deleted).to_bytes(8, "big"),
                *(
                    b"".join([len(name).to_bytes(8, "big"), name.encode("utf-8")])
                    for name in self.deleted
                ),
            ]
        )

    @classmethod
    def frombuffer(cls, data: bytes) -> "DatasetDiff":
        """Creates a DatasetDiff object from bytes"""
        dataset_diff = cls()
        dataset_diff.info_updated = bool(int.from_bytes(data[:1], "big"))
        len_renamed = int.from_bytes(data[1:9], "big")
        pos = 9
        for _ in range(len_renamed):
            len_old, len_new = int.from_bytes(
                data[pos : pos + 8], "big"
            ), int.from_bytes(data[pos + 8 : pos + 16], "big")
            pos += 16
            old_name, new_name = data[pos : pos + len_old].decode("utf-8"), data[
                pos + len_old : pos + len_old + len_new
            ].decode("utf-8")
            pos += len_old + len_new
            dataset_diff.renamed[old_name] = new_name
        len_deleted = int.from_bytes(data[pos : pos + 8], "big")
        pos += 8
        for _ in range(len_deleted):
            len_name = int.from_bytes(data[pos : pos + 8], "big")
            pos += 8
            name = data[pos : pos + len_name].decode("utf-8")
            pos += len_name
            dataset_diff.deleted.append(name)
        return dataset_diff

    @property
    def nbytes(self):
        """Returns number of bytes required to store the dataset diff"""
        return 1

    def modify_info(self) -> None:
        """Stores information that the info has changed"""
        self.info_updated = True
        self.is_dirty = True

    def tensor_renamed(self, old_name, new_name):
        """Adds old and new name of a tensor that was renamed to renamed"""
        for old, new in self.renamed.items():
            if old_name == new:
                if old == new_name:
                    self.renamed.pop(old)
                else:
                    self.renamed[old] = new_name
                break
        else:
            self.renamed[old_name] = new_name

        self.is_dirty = True

    def tensor_deleted(self, name):
        """Adds name of deleted tensor to deleted"""
        if name not in self.deleted:
            for old, new in self.renamed.items():
                if name == new:
                    self.renamed.pop(old)
                    self.deleted.append(old)
                    break
            else:
                self.deleted.append(name)
            self.is_dirty = True


def load_dataset_diff(dataset):
    storage: LRUCache = dataset.storage
    path = get_dataset_diff_key(dataset.version_state["commit_id"])
    try:
        diff = storage.get_hub_object(path, DatasetDiff)
    except KeyError:
        diff = DatasetDiff()
    storage.register_hub_object(path, diff)
    return diff



from typing import Optional
from hub.client.client import HubBackendClient
from hub.util.path import is_hub_cloud_path


def remove_path_from_backend(path: str, token: Optional[str] = None) -> None:
    if is_hub_cloud_path(path):
        client = HubBackendClient(token=token)
        split_path = path.split("/")
        org_id, ds_name = split_path[2], split_path[3]
        client.delete_dataset_entry(org_id, ds_name)



from typing import Union
from hub.util.exceptions import InvalidOperationError
from typing import Callable
from functools import wraps
import hub


def invalid_view_op(callable: Callable):
    @wraps(callable)
    def inner(x, *args, **kwargs):
        ds = x if isinstance(x, hub.Dataset) else x.dataset
        if not ds.__dict__.get("_allow_view_updates"):
            is_del = callable.__name__ == "delete"
            managed_view = "_view_entry" in ds.__dict__
            has_vds = "_vds" in ds.__dict__
            is_view = not x.index.is_trivial() or has_vds or managed_view
            if is_view and not (is_del and (has_vds or managed_view)):
                raise InvalidOperationError(
                    callable.__name__,
                    type(x).__name__,
                )
        return callable(x, *args, **kwargs)

    return inner



from hub.client.log import logger
from hub.util.path import is_hub_cloud_path, get_org_id_and_ds_name


def log_visualizer_link(ds_path, source_ds_url=""):
    msg = "This dataset can be visualized in Jupyter Notebook by ds.visualize()"
    if is_hub_cloud_path(ds_path):
        org_id, ds_name = get_org_id_and_ds_name(ds_path)
        url = f"https://app.activeloop.ai/{org_id}/{ds_name}"
        if url.endswith("/queries"):  # Ignore user queries ds
            pass
        elif "/.queries/" in url:  # Is a view
            if "/queries/" in url:  # Stored in user queries ds
                if is_hub_cloud_path(source_ds_url):
                    org_id, ds_name = get_org_id_and_ds_name(source_ds_url)
                    source_ds_url = f"https://app.activeloop.ai/{org_id}/{ds_name}"
                    view_id = url.split("/.queries/", 1)[1]
                    if view_id.endswith("_OPTIMIZED"):
                        view_id = view_id[: -len("_OPTIMIZED")]
                    view_id = view_id[len(f"[{org_id}][{ds_name}]") :]
                    url = source_ds_url + "?view=" + view_id
                    logger.info(msg + " or at " + url)
                else:
                    logger.info(msg + ".")
            else:  # Stored in ds
                ds_url, view_id = url.split("/.queries/", 1)
                if view_id.endswith("_OPTIMIZED"):
                    view_id = view_id[: -len("_OPTIMIZED")]
                url = ds_url + "?view=" + view_id
                logger.info(msg + " or at " + url)
        else:
            logger.info(msg + " or at " + url)
    else:
        logger.info(msg + ".")



import hub
import requests
from typing import Optional
from hub.util.exceptions import (
    AgreementNotAcceptedError,
    AuthorizationException,
    LoginException,
    InvalidPasswordException,
    ManagedCredentialsNotFoundError,
    NotLoggedInAgreementError,
    ResourceNotFoundException,
    InvalidTokenException,
    UserNotLoggedInException,
    TokenPermissionError,
)
from hub.client.utils import check_response_status, write_token, read_token
from hub.client.config import (
    ACCEPT_AGREEMENTS_SUFFIX,
    REJECT_AGREEMENTS_SUFFIX,
    GET_MANAGED_CREDS_SUFFIX,
    HUB_REST_ENDPOINT,
    HUB_REST_ENDPOINT_LOCAL,
    HUB_REST_ENDPOINT_DEV,
    GET_TOKEN_SUFFIX,
    HUB_REST_ENDPOINT_STAGING,
    REGISTER_USER_SUFFIX,
    DEFAULT_REQUEST_TIMEOUT,
    GET_DATASET_CREDENTIALS_SUFFIX,
    CREATE_DATASET_SUFFIX,
    DATASET_SUFFIX,
    LIST_DATASETS,
    GET_USER_PROFILE,
    SEND_EVENT_SUFFIX,
    UPDATE_SUFFIX,
    GET_PRESIGNED_URL_SUFFIX,
)
from hub.client.log import logger
import jwt  # should add it to requirements.txt

# for these codes, we will retry requests upto 3 times
retry_status_codes = {502}


class HubBackendClient:
    """Communicates with Activeloop Backend"""

    def __init__(self, token: Optional[str] = None):
        self.version = hub.__version__
        self.auth_header = None
        if token is None:
            self.token = self.get_token()
        else:
            self.token = token
        self.auth_header = f"Bearer {self.token}"

    def get_token(self):
        """Returns a token"""
        token = read_token()
        if token is None:
            token = self.request_auth_token(username="public", password="")
            write_token(token)

        return token

    def request(
        self,
        method: str,
        relative_url: str,
        endpoint: Optional[str] = None,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        files: Optional[dict] = None,
        json: Optional[dict] = None,
        headers: Optional[dict] = None,
        timeout: Optional[int] = DEFAULT_REQUEST_TIMEOUT,
    ):
        """Sends a request to the backend.

        Args:
            method (str): The method for sending the request. Should be one of 'GET', 'OPTIONS', 'HEAD', 'POST', 'PUT',
                'PATCH', or 'DELETE'.
            relative_url (str): The suffix to be appended to the end of the endpoint url.
            endpoint(str, optional): The endpoint to send the request to.
            params (dict, optional): Dictionary to send in the query string for the request.
            data (dict, optional): Dictionary to send in the body of the request.
            files (dict, optional): Dictionary of 'name': file-like-objects (or {'name': file-tuple}) for multipart
                encoding upload.
                file-tuple can be a 2-tuple (filename, fileobj), 3-tuple (filename, fileobj, content_type)
                or a 4-tuple (filename, fileobj, content_type, custom_headers), where 'content-type' is a string
                defining the content type of the given file and 'custom_headers' a dict-like object containing
                additional headers to add for the file.
            json (dict, optional): A JSON serializable Python object to send in the body of the request.
            headers (dict, optional): Dictionary of HTTP Headers to send with the request.
            timeout (float,optional): How many seconds to wait for the server to send data before giving up.

        Raises:
            InvalidPasswordException: `password` cannot be `None` inside `json`.

        Returns:
            requests.Response: The response received from the server.
        """
        params = params or {}
        data = data or None
        files = files or None
        json = json or None
        endpoint = endpoint or self.endpoint()
        endpoint = endpoint.strip("/")
        relative_url = relative_url.strip("/")
        request_url = f"{endpoint}/{relative_url}"
        headers = headers or {}
        headers["hub-cli-version"] = self.version
        headers["Authorization"] = self.auth_header

        # clearer error than `ServerUnderMaintenence`
        if json is not None and "password" in json and json["password"] is None:
            # do NOT pass in the password here. `None` is explicitly typed.
            raise InvalidPasswordException("Password cannot be `None`.")

        status_code = None
        tries = 0
        while status_code is None or (status_code in retry_status_codes and tries < 3):
            response = requests.request(
                method,
                request_url,
                params=params,
                data=data,
                json=json,
                headers=headers,
                files=files,
                timeout=timeout,
            )
            status_code = response.status_code
            tries += 1
        check_response_status(response)
        return response

    def endpoint(self):
        if hub.client.config.USE_LOCAL_HOST:
            return HUB_REST_ENDPOINT_LOCAL
        if hub.client.config.USE_DEV_ENVIRONMENT:
            return HUB_REST_ENDPOINT_DEV
        if hub.client.config.USE_STAGING_ENVIRONMENT:
            return HUB_REST_ENDPOINT_STAGING

        return HUB_REST_ENDPOINT

    def request_auth_token(self, username: str, password: str):
        """Sends a request to backend to retrieve auth token.

        Args:
            username (str): The Activeloop username to request token for.
            password (str): The password of the account.

        Returns:
            string: The auth token corresponding to the accound.

        Raises:
            UserNotLoggedInException: if user is not authorised
            LoginException: If there is an issue retrieving the auth token.

        """
        json = {"username": username, "password": password}
        response = self.request("POST", GET_TOKEN_SUFFIX, json=json)

        try:
            token_dict = response.json()
            token = token_dict["token"]
        except Exception:
            raise LoginException()
        return token

    def send_register_request(self, username: str, email: str, password: str):
        """Sends a request to backend to register a new user.

        Args:
            username (str): The Activeloop username to create account for.
            email (str): The email id to link with the Activeloop account.
            password (str): The new password of the account. Should be atleast 6 characters long.
        """

        json = {"username": username, "email": email, "password": password}
        self.request("POST", REGISTER_USER_SUFFIX, json=json)

    def get_dataset_credentials(
        self,
        org_id: str,
        ds_name: str,
        mode: Optional[str] = None,
        no_cache: bool = False,
    ):
        """Retrieves temporary 12 hour credentials for the required dataset from the backend.

        Args:
            org_id (str): The name of the user/organization to which the dataset belongs.
            ds_name (str): The name of the dataset being accessed.
            mode (str, optional): The mode in which the user has requested to open the dataset.
                If not provided, the backend will set mode to 'a' if user has write permission, else 'r'.
            no_cache (bool): If True, cached creds are ignored and new creds are returned. Default False.

        Returns:
            tuple: containing full url to dataset, credentials, mode and expiration time respectively.

        Raises:
            UserNotLoggedInException: When user is not logged in
            InvalidTokenException: If the specified token is invalid
            TokenPermissionError: when there are permission or other errors related to token
            AgreementNotAcceptedError: when user has not accepted the agreement
            NotLoggedInAgreementError: when user is not logged in and dataset has agreement which needs to be signed
        """
        relative_url = GET_DATASET_CREDENTIALS_SUFFIX.format(org_id, ds_name)
        try:
            response = self.request(
                "GET",
                relative_url,
                endpoint=self.endpoint(),
                params={"mode": mode, "no_cache": no_cache},
            ).json()
        except Exception as e:
            if isinstance(e, AuthorizationException):
                authorization_exception_prompt = "You don't have permission to "
                response_data = e.response.json()
                code = response_data.get("code")
                if code == 1:
                    agreements = response_data["agreements"]
                    agreements = [agreement["text"] for agreement in agreements]
                    raise AgreementNotAcceptedError(agreements) from e
                elif code == 2:
                    raise NotLoggedInAgreementError from e
                else:
                    try:
                        decoded_token = jwt.decode(
                            self.token, options={"verify_signature": False}
                        )
                    except Exception:
                        raise InvalidTokenException

                    if (
                        authorization_exception_prompt in response_data["description"]
                        and decoded_token["id"] == "public"
                    ):
                        raise UserNotLoggedInException()
                    raise TokenPermissionError()
            raise

        full_url = response.get("path")
        creds = response["creds"]
        mode = response["mode"]
        expiration = creds["expiration"]
        return full_url, creds, mode, expiration

    def send_event(self, event_json: dict):
        """Sends an event to the backend.

        Args:
            event_json (dict): The event to be sent.
        """
        self.request("POST", SEND_EVENT_SUFFIX, json=event_json)

    def create_dataset_entry(self, username, dataset_name, meta, public=True):
        tag = f"{username}/{dataset_name}"
        repo = f"protected/{username}"

        response = self.request(
            "POST",
            CREATE_DATASET_SUFFIX,
            json={
                "tag": tag,
                "repository": repo,
                "public": public,
                "rewrite": True,
                "meta": meta,
            },
            endpoint=self.endpoint(),
        )

        if response.status_code == 200:
            logger.info("Your Hub dataset has been successfully created!")
            if public is False:
                logger.info("The dataset is private so make sure you are logged in!")

    def get_managed_creds(self, org_id, creds_key):
        """Retrieves the managed credentials for the given org_id and creds_key.

        Args:
            org_id (str): The name of the user/organization to which the dataset belongs.
            creds_key (str): The key corresponding to the managed credentials.

        Returns:
            dict: The managed credentials.

        Raises:
            ManagedCredentialsNotFoundError: If the managed credentials do not exist for the given organization.
        """
        relative_url = GET_MANAGED_CREDS_SUFFIX.format(org_id)
        try:
            resp = self.request(
                "GET",
                relative_url,
                endpoint=self.endpoint(),
                params={"query": creds_key},
            ).json()
        except ResourceNotFoundException:
            raise ManagedCredentialsNotFoundError(org_id, creds_key) from None
        creds = resp["creds"]
        key_mapping = {
            "access_key": "aws_access_key_id",
            "secret_key": "aws_secret_access_key",
            "session_token": "aws_session_token",
            "token": "aws_session_token",
            "region": "aws_region",
        }
        final_creds = {}
        for key, value in creds.items():
            if key == "access_token":
                key = "Authorization"
                value = f"Bearer {value}"
            elif key in key_mapping:
                key = key_mapping[key]
            final_creds[key] = value
        return final_creds

    def delete_dataset_entry(self, username, dataset_name):
        tag = f"{username}/{dataset_name}"
        suffix = f"{DATASET_SUFFIX}/{tag}"
        self.request(
            "DELETE",
            suffix,
            endpoint=self.endpoint(),
        ).json()

    def accept_agreements(self, org_id, ds_name):
        """Accepts the agreements for the given org_id and ds_name.

        Args:
            org_id (str): The name of the user/organization to which the dataset belongs.
            ds_name (str): The name of the dataset being accessed.
        """
        relative_url = ACCEPT_AGREEMENTS_SUFFIX.format(org_id, ds_name)
        self.request(
            "POST",
            relative_url,
            endpoint=self.endpoint(),
        ).json()

    def reject_agreements(self, org_id, ds_name):
        """Rejects the agreements for the given org_id and ds_name.

        Args:
            org_id (str): The name of the user/organization to which the dataset belongs.
            ds_name (str): The name of the dataset being accessed.
        """
        relative_url = REJECT_AGREEMENTS_SUFFIX.format(org_id, ds_name)
        self.request(
            "POST",
            relative_url,
            endpoint=self.endpoint(),
        ).json()

    def rename_dataset_entry(self, username, old_name, new_name):
        suffix = UPDATE_SUFFIX.format(username, old_name)
        self.request(
            "PUT", suffix, endpoint=self.endpoint(), json={"basename": new_name}
        )

    def get_user_organizations(self):
        """Get list of user organizations from the backend. If user is not logged in, returns ['public'].

        Returns:
            list: user/organization names
        """
        response = self.request(
            "GET", GET_USER_PROFILE, endpoint=self.endpoint()
        ).json()
        return response["organizations"]

    def get_workspace_datasets(
        self, workspace: str, suffix_public: str, suffix_user: str
    ):
        organizations = self.get_user_organizations()
        if workspace in organizations:
            response = self.request(
                "GET",
                suffix_user,
                endpoint=self.endpoint(),
                params={"organization": workspace},
            ).json()
        else:
            print(
                f'You are not a member of organization "{workspace}". List of accessible datasets from "{workspace}": ',
            )
            response = self.request(
                "GET",
                suffix_public,
                endpoint=self.endpoint(),
                params={"organization": workspace},
            ).json()
        return response

    def get_datasets(self, workspace: str):
        suffix_public = LIST_DATASETS.format("public")
        suffix_user = LIST_DATASETS.format("all")
        if workspace:
            res_datasets = self.get_workspace_datasets(
                workspace, suffix_public, suffix_user
            )
        else:
            public_datasets = self.request(
                "GET",
                suffix_public,
                endpoint=self.endpoint(),
            ).json()
            user_datasets = self.request(
                "GET",
                suffix_user,
                endpoint=self.endpoint(),
            ).json()
            res_datasets = public_datasets + user_datasets
        return [ds["_id"] for ds in res_datasets]

    def update_privacy(self, username: str, dataset_name: str, public: bool):
        suffix = UPDATE_SUFFIX.format(username, dataset_name)
        self.request("PUT", suffix, endpoint=self.endpoint(), json={"public": public})

    def get_presigned_url(self, org_id, ds_id, chunk_path, expiration=3600):
        relative_url = GET_PRESIGNED_URL_SUFFIX.format(org_id, ds_id)
        response = self.request(
            "GET",
            relative_url,
            endpoint=self.endpoint(),
            params={"chunk_path": chunk_path, "expiration": expiration},
        ).json()
        presigned_url = response["data"]
        return presigned_url

    def get_user_profile(self):
        response = self.request(
            "GET",
            "/api/user/profile",
            endpoint=self.endpoint(),
        )
        return response.json()



def normalize_index(index, nframes):
    reverse = False
    if index is None:
        start = 0
        stop = nframes
        step = 1
    elif isinstance(index, int):
        if index >= 0:
            start = index
        else:
            start = nframes + index
        stop = start + 1
        step = 1
    elif isinstance(index, slice):
        step = index.step
        if step is None:
            step = 1
        elif step < 0:
            step = abs(step)
            reverse = True

        start = index.start
        if start is None:
            start = 0 if not reverse else nframes
        elif start < 0:
            start = nframes + start

        stop = index.stop
        if stop is None:
            stop = nframes if not reverse else -1
        elif stop < 0:
            stop = nframes + stop

        if reverse:
            start, stop = stop + 1, start + 1

    return start, stop, step, reverse



import itertools


"""
Supported compressions (formats):

    Image : bmp, dib, gif, ico, jpeg, jpeg2000, pcx, png, ppm, sgi, tga, tiff, webp, wmf, xbm
    Audio : flac, mp3, wav
    Video : mp4, mkv, avi
    Bytes : lz4

__Note__:- 

For video compressions, we only support already compressed data read using hub.read. We do not actually compress the video data. 

Also, when using hub.read with one of the video compressions, ensure that the compression matches, otherwise hub will be unable to compress the data to the specified compression.

"""
from PIL import Image  # type: ignore


BYTE_COMPRESSIONS = [
    "lz4",
]


IMAGE_COMPRESSIONS = [
    "bmp",
    "dib",
    "gif",
    "ico",
    "jpeg",
    "jpeg2000",
    "pcx",
    "png",
    "ppm",
    "sgi",
    "tga",
    "tiff",
    "webp",
    "wmf",
    "xbm",
]

IMAGE_COMPRESSION_EXT_DICT = {
    "apng": [".png"],
    "bmp": [".bmp"],
    "dib": [".dib"],
    "gif": [".gif"],
    "ico": [".ico"],
    "jpeg": [".jpg", ".jpeg", ".jfif", ".pjpeg", ".pjp"],
    "jpeg2000": [
        ".jp2",
        ".j2k",
        ".jpf",
        ".jpm",
        ".jpg2",
        ".j2c",
        ".jpc",
        ".jpx",
        ".mj2",
    ],
    "pcx": [".pcx"],
    "png": [".png"],
    "ppm": [".pbm", ".pgm", ".ppm", ".pnm"],
    "sgi": [".sgi"],
    "tga": [".tga"],
    "tiff": [".tiff", ".tif"],
    "webp": [".webp"],
    "wmf": [".wmf"],
    "xbm": [".xbm"],
}


IMAGE_COMPRESSION_EXTENSIONS = list(
    set(itertools.chain(*IMAGE_COMPRESSION_EXT_DICT.values()))
)

VIDEO_COMPRESSIONS = ["mp4", "mkv", "avi"]

AUDIO_COMPRESSIONS = ["mp3", "flac", "wav"]

POINT_CLOUD_COMPRESSIONS = ["las"]


# Just constants
BYTE_COMPRESSION = "byte"
IMAGE_COMPRESSION = "image"
VIDEO_COMPRESSION = "video"
AUDIO_COMPRESSION = "audio"
POINT_CLOUD_COMPRESSION = "point_cloud"


COMPRESSION_TYPES = [
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    AUDIO_COMPRESSION,
    VIDEO_COMPRESSION,
    POINT_CLOUD_COMPRESSION,
]

# Pillow plugins for some formats might not be installed:
Image.init()
IMAGE_COMPRESSIONS = [
    c for c in IMAGE_COMPRESSIONS if c.upper() in Image.SAVE and c.upper() in Image.OPEN
]


IMAGE_COMPRESSIONS.insert(0, "apng")
IMAGE_COMPRESSIONS.insert(2, "dcm")

SUPPORTED_COMPRESSIONS = [
    *BYTE_COMPRESSIONS,
    *IMAGE_COMPRESSIONS,
    *AUDIO_COMPRESSIONS,
    *VIDEO_COMPRESSIONS,
    *POINT_CLOUD_COMPRESSIONS,
]
SUPPORTED_COMPRESSIONS = list(sorted(set(SUPPORTED_COMPRESSIONS)))  # type: ignore
SUPPORTED_COMPRESSIONS.append(None)  # type: ignore

COMPRESSION_ALIASES = {"jpg": "jpeg", "tif": "tiff", "jp2": "jpeg2000"}

# If `True`  compression format has to be the same between samples in the same tensor.
# If `False` compression format can   be different between samples in the same tensor.
USE_UNIFORM_COMPRESSION_PER_SAMPLE = True


_compression_types = {}
for c in IMAGE_COMPRESSIONS:
    _compression_types[c] = IMAGE_COMPRESSION
for c in BYTE_COMPRESSIONS:
    _compression_types[c] = BYTE_COMPRESSION
for c in VIDEO_COMPRESSIONS:
    _compression_types[c] = VIDEO_COMPRESSION
for c in AUDIO_COMPRESSIONS:
    _compression_types[c] = AUDIO_COMPRESSION
for c in POINT_CLOUD_COMPRESSIONS:
    _compression_types[c] = POINT_CLOUD_COMPRESSION


def get_compression_type(c):
    """Returns the compression type for the given compression name."""
    if c is None:
        return None
    ret = _compression_types.get(c)
    if ret is None and c.upper() in Image.OPEN:
        ret = IMAGE_COMPRESSION
    if ret is None:
        raise KeyError(c)
    return ret



"""
"htype" is the class of a tensor: image, bounding box, generic tensor, etc.

When not specified, the unspecified options will be inferred from the data:

>>> ds.create_tensor("my_tensor")
>>> ds.my_tensor.append(1)
>>> ds.my_tensor.dtype
int64

If you know beforehand, you can use htype at creation:

>>> ds.create_tensor("my_tensor", htype="image", sample_compression=None)

Specifying an htype allows for strict settings and error handling, and it is critical for increasing the performance of hub datasets containing rich data such as images and videos.

Supported htypes and their respective defaults are:

+----------------+-----------+---------------+
| HTYPE          |  DTYPE    |  COMPRESSION  |
+================+===========+===============+
| image          |  uint8    |  None         |
+----------------+-----------+---------------+
| image.rgb      |  uint8    |  None         |
+----------------+-----------+---------------+
| image.gray     |  uint8    |  None         |
+----------------+-----------+---------------+
| class_label    |  uint32   |  None         |
+----------------+-----------+---------------+
| bbox           |  float32  |  None         |
+----------------+-----------+---------------+
| video          |  uint8    |  None         |
+----------------+-----------+---------------+
| binary_mask    |  bool     |  None         |
+----------------+-----------+---------------+
| segment_mask   |  uint32   |  None         |
+----------------+-----------+---------------+
| keypoints_coco |  int32    |  None         |
+----------------+-----------+---------------+
| point          |  int32    |  None         |
+----------------+-----------+---------------+
| audio          |  float64  |  None         |
+----------------+-----------+---------------+
| text           |  str      |  None         |
+----------------+-----------+---------------+
| json           |  Any      |  None         |
+----------------+-----------+---------------+
| list           |  List     |  None         |
+----------------+-----------+---------------+
| dicom          |  None     |  dcm          |
+----------------+-----------+---------------+
| link           |  str      |  None         |
+----------------+-----------+---------------+
| sequence       |  None     |  None         |
+----------------+-----------+---------------+

Sequence htype
~~~~~~~~~~~~~~

- A special meta htype for tensors where each sample is a sequence. The items in the sequence are samples of another htype.
- It is a wrapper htype that can wrap other htypes like ``sequence[image]``, ``sequence[video]``, ``sequence[text]``, etc.

Examples
--------

>>> ds.create_tensor("seq", htype="sequence")
>>> ds.seq.append([1, 2, 3])
>>> ds.seq.append([4, 5, 6])
>>> ds.seq.numpy()
array([[[1],
        [2],
        [3]],
       [[4],
        [5],
        [6]]])

>>> ds.create_tensor("image_seq", htype="sequence[image]", sample_compression="jpg")
>>> ds.image_seq.append([hub.read("img01.jpg"), hub.read("img02.jpg")])

Link htype
~~~~~~~~~~

- Link htype is a special meta htype that allows linking of external data (files) to the dataset, without storing the data in the dataset itself.
- Moreover, there can be variations in this htype, such as ``link[image]``, ``link[video]``, ``link[audio]``, etc. that would enable the activeloop visualizer to correctly display the data.
- No data is actually loaded until you try to read the sample from a dataset.
- There are a few exceptions to this:-
    - If ``verify=True`` was specified during ``create_tensor`` of the tensor to which this is being added, some metadata is read to verify the integrity of the sample.
    - If ``create_shape_tensor=True`` was specified during ``create_tensor`` of the tensor to which this is being added, the shape of the sample is read.
    - If ``create_sample_info_tensor=True`` was specified during ``create_tensor`` of the tensor to which this is being added, the sample info is read.

Examples
--------

>>> ds = hub.dataset("......")

Add the names of the creds you want to use (not needed for http/local urls)

>>> ds.add_creds_key("MY_S3_KEY")
>>> ds.add_creds_key("GCS_KEY")

Populate the names added with creds dictionary
These creds are only present temporarily and will have to be repopulated on every reload

>>> ds.populate_creds("MY_S3_KEY", {})   # add creds here
>>> ds.populate_creds("GCS_KEY", {})    # add creds here

Create a tensor that can contain links

>>> ds.create_tensor("img", htype="link[image]", verify=True, create_shape_tensor=False, create_sample_info_tensor=False)

Populate the tensor with links

>>> ds.img.append(hub.link("s3://abc/def.jpeg", creds_key="MY_S3_KEY"))
>>> ds.img.append(hub.link("gcs://ghi/jkl.png", creds_key="GCS_KEY"))
>>> ds.img.append(hub.link("https://picsum.photos/200/300")) # http path doesn’t need creds
>>> ds.img.append(hub.link("./path/to/cat.jpeg")) # local path doesn’t need creds
>>> ds.img.append(hub.link("s3://abc/def.jpeg"))  # this will throw an exception as cloud paths always need creds_key
>>> ds.img.append(hub.link("s3://abc/def.jpeg", creds_key="ENV"))  # this will use creds from environment

Accessing the data

>>> for i in range(5):
...     ds.img[i].numpy()
...

Updating a sample

>>> ds.img[0] = hub.link("./data/cat.jpeg")

"""

from typing import Dict
from hub.compression import (
    IMAGE_COMPRESSIONS,
    VIDEO_COMPRESSIONS,
    AUDIO_COMPRESSIONS,
    BYTE_COMPRESSIONS,
    COMPRESSION_ALIASES,
    POINT_CLOUD_COMPRESSIONS,
)


class htype:
    DEFAULT = "generic"
    IMAGE = "image"
    IMAGE_RGB = "image.rgb"
    IMAGE_GRAY = "image.gray"
    CLASS_LABEL = "class_label"
    BBOX = "bbox"
    BBOX_3D = "bbox.3d"
    VIDEO = "video"
    BINARY_MASK = "binary_mask"
    INSTANCE_LABEL = "instance_label"
    SEGMENT_MASK = "segment_mask"
    KEYPOINTS_COCO = "keypoints_coco"
    POINT = "point"
    AUDIO = "audio"
    TEXT = "text"
    JSON = "json"
    LIST = "list"
    DICOM = "dicom"
    POINT_CLOUD = "point_cloud"
    POINT_CLOUD_CALIBRATION_MATRIX = "point_cloud.calibration_matrix"


# used for requiring the user to specify a value for htype properties. notates that the htype property has no default.
REQUIRE_USER_SPECIFICATION = "require_user_specification"

# used for `REQUIRE_USER_SPECIFICATION` enforcement. this should be used instead of `None` for default user method arguments.
UNSPECIFIED = "unspecified"


HTYPE_CONFIGURATIONS: Dict[str, Dict] = {
    htype.DEFAULT: {"dtype": None},
    htype.IMAGE: {
        "dtype": "uint8",
    },
    htype.IMAGE_RGB: {
        "dtype": "uint8",
    },
    htype.IMAGE_GRAY: {
        "dtype": "uint8",
    },
    htype.CLASS_LABEL: {
        "dtype": "uint32",
        "class_names": [],
        "_info": ["class_names"],  # class_names should be stored in info, not meta
        "_disable_temp_transform": False,
    },
    htype.BBOX: {"dtype": "float32", "coords": {}, "_info": ["coords"]},
    htype.BBOX_3D: {"dtype": "float32", "coords": {}, "_info": ["coords"]},
    htype.AUDIO: {"dtype": "float64"},
    htype.VIDEO: {"dtype": "uint8"},
    htype.BINARY_MASK: {
        "dtype": "bool"
    },  # TODO: pack numpy arrays to store bools as 1 bit instead of 1 byte
    htype.INSTANCE_LABEL: {"dtype": "uint32"},
    htype.SEGMENT_MASK: {
        "dtype": "uint32",
        "class_names": [],
        "_info": ["class_names"],
    },
    htype.KEYPOINTS_COCO: {"dtype": "int32"},
    htype.POINT: {"dtype": "int32"},
    htype.JSON: {
        "dtype": "Any",
    },
    htype.LIST: {"dtype": "List"},
    htype.TEXT: {"dtype": "str"},
    htype.DICOM: {"sample_compression": "dcm"},
    htype.POINT_CLOUD: {"sample_compression": "las"},
    htype.POINT_CLOUD_CALIBRATION_MATRIX: {"dtype": "float32"},
}

HTYPE_VERIFICATIONS: Dict[str, Dict] = {
    htype.BBOX: {"coords": {"type": dict, "keys": ["type", "mode"]}}
}

_image_compressions = (
    IMAGE_COMPRESSIONS[:] + BYTE_COMPRESSIONS + list(COMPRESSION_ALIASES)
)
_image_compressions.remove("dcm")

HTYPE_SUPPORTED_COMPRESSIONS = {
    htype.IMAGE: _image_compressions,
    htype.IMAGE_RGB: _image_compressions,
    htype.IMAGE_GRAY: _image_compressions,
    htype.VIDEO: VIDEO_COMPRESSIONS[:],
    htype.AUDIO: AUDIO_COMPRESSIONS[:],
    htype.TEXT: BYTE_COMPRESSIONS[:],
    htype.LIST: BYTE_COMPRESSIONS[:],
    htype.JSON: BYTE_COMPRESSIONS[:],
    htype.POINT_CLOUD: POINT_CLOUD_COMPRESSIONS[:],
    htype.DICOM: ["dcm"],
}


# these configs are added to every `htype`
COMMON_CONFIGS = {
    "sample_compression": None,
    "chunk_compression": None,
    "dtype": None,
    "max_chunk_size": None,
    "tiling_threshold": None,
    "is_sequence": False,
    "is_link": False,
    "hidden": False,
    "links": None,
    "verify": False,
}


for config in HTYPE_CONFIGURATIONS.values():
    for key, v in COMMON_CONFIGS.items():
        # only update if not specified explicitly
        if key not in config:
            config[key] = v


def verify_htype_key_value(htype, key, value):
    htype_verifications = HTYPE_VERIFICATIONS.get(htype, {})
    if key in htype_verifications:
        expected_type = htype_verifications[key].get("type")
        if expected_type and not isinstance(value, expected_type):
            raise TypeError(f"{key} must be of type {expected_type}, not {type(value)}")
        if expected_type == dict:
            expected_keys = set(htype_verifications[key].get("keys"))
            present_keys = set(value.keys())
            if expected_keys and not present_keys.issubset(expected_keys):
                raise KeyError(f"{key} must have keys belong to {expected_keys}")



from hub.compression import (
    BYTE_COMPRESSION,
    VIDEO_COMPRESSION,
    AUDIO_COMPRESSION,
    get_compression_type,
)
from hub.core.fast_forwarding import version_compare
from hub.core.tiling.sample_tiles import SampleTiles
from hub.core.partial_sample import PartialSample
from hub.util.compression import get_compression_ratio  # type: ignore
from hub.util.exceptions import TensorInvalidSampleShapeError
from hub.util.casting import intelligent_cast
from hub.util.json import HubJsonDecoder, HubJsonEncoder, validate_json_object
from hub.core.sample import Sample, SampleValue  # type: ignore
from hub.core.compression import compress_array, compress_bytes
from typing import Optional, Sequence, Union, Tuple
import hub
import numpy as np
import struct
import json
from urllib.request import Request, urlopen

BaseTypes = Union[np.ndarray, list, int, float, bool, np.integer, np.floating, np.bool_]
HEADER_SIZE_BYTES = 13


def infer_header_num_bytes(
    version: str, shape_info: np.ndarray, byte_positions: np.ndarray
):
    """Calculates the number of header bytes in a chunk without serializing it.

    Args:
        version: (str) Version of hub library
        shape_info: (numpy.ndarray) Encoded shapes info from the chunk's `ShapeEncoder` instance.
        byte_positions: (numpy.ndarray) Encoded byte positions from the chunk's `BytePositionsEncoder` instance.

    Returns:
        Length of the headers of chunk when serialized as int"""
    return len(version) + shape_info.nbytes + byte_positions.nbytes + HEADER_SIZE_BYTES


def infer_chunk_num_bytes(
    version: str,
    shape_info: np.ndarray,
    byte_positions: np.ndarray,
    data: Optional[Union[Sequence[bytes], Sequence[memoryview]]] = None,
    len_data: Optional[int] = None,
) -> int:
    """Calculates the number of bytes in a chunk without serializing it. Used by `LRUCache` to determine if a chunk can be cached.

    Args:
        version: (str) Version of hub library
        shape_info: (numpy.ndarray) Encoded shapes info from the chunk's `ShapeEncoder` instance.
        byte_positions: (numpy.ndarray) Encoded byte positions from the chunk's `BytePositionsEncoder` instance.
        data: (list) `_data` field of the chunk
        len_data: (int, optional) Number of bytes in the chunk

    Returns:
        Length of the chunk when serialized as int
    """
    # NOTE: Assumption: version string contains ascii characters only (ord(c) < 128)
    # NOTE: Assumption: len(version) < 256
    if len_data is None:
        len_data = sum(map(len, data))  # type: ignore

    header_size = infer_header_num_bytes(version, shape_info, byte_positions)
    return header_size + len_data


def serialize_chunk(
    version: str,
    shape_info: np.ndarray,
    byte_positions: np.ndarray,
    data: Union[Sequence[bytes], Sequence[memoryview]],
    len_data: Optional[int] = None,
) -> memoryview:
    """Serializes a chunk's headers and data into a single byte stream. This is how the chunk will be written to the storage provider.

    Args:
        version: (str) Version of hub library.
        shape_info: (numpy.ndarray) Encoded shapes info from the chunk's `ShapeEncoder` instance.
        byte_positions: (numpy.ndarray) Encoded byte positions from the chunk's `BytePositionsEncoder` instance.
        data: (list) `_data` field of the chunk.
        len_data: (int, optional) Number of bytes in the chunk.

    Returns:
        Serialized chunk as memoryview.
    """
    nbytes = infer_chunk_num_bytes(version, shape_info, byte_positions, data, len_data)
    flatbuff = bytearray(nbytes)
    offset = write_version(version, flatbuff)
    offset = write_shape_info(shape_info, flatbuff, offset)
    offset = write_byte_positions(byte_positions, flatbuff, offset)
    offset = write_actual_data(data, flatbuff, offset)
    return memoryview(flatbuff)


def write_version(version, buffer) -> int:
    """Writes version info to the buffer, returns offset."""
    len_version = len(version)
    buffer[0] = len_version
    buffer[1 : 1 + len_version] = version.encode("ascii")
    offset = 1 + len_version
    return offset


def write_shape_info(shape_info, buffer, offset) -> int:
    """Writes shape info to the buffer, takes offset into account and returns updated offset."""
    if shape_info.ndim == 1:
        offset += 8
    else:
        buffer[offset : offset + 8] = struct.pack("<ii", *shape_info.shape)
        offset += 8

        buffer[offset : offset + shape_info.nbytes] = shape_info.tobytes()
        offset += shape_info.nbytes
    return offset


def write_byte_positions(byte_positions, buffer, offset) -> int:
    """Writes byte positions info to the buffer, takes offset into account and returns updated offset."""
    if byte_positions.ndim == 1:
        offset += 4
    else:
        buffer[offset : offset + 4] = byte_positions.shape[0].to_bytes(4, "little")
        offset += 4

        buffer[offset : offset + byte_positions.nbytes] = byte_positions.tobytes()
        offset += byte_positions.nbytes
    return offset


def write_actual_data(data, buffer, offset) -> int:
    """Writes actual chunk data to the buffer, takes offset into account and returns updated offset"""
    for byts in data:
        n = len(byts)
        buffer[offset : offset + n] = byts
        offset += n
    return offset


def get_header_from_url(url: str):
    # Note: to be only used for chunks contains a single sample
    enc_dtype = np.dtype(hub.constants.ENCODING_DTYPE)
    itemsize = enc_dtype.itemsize

    headers = {"Range": "bytes=0-100"}

    request = Request(url, None, headers)
    byts = urlopen(request).read()

    len_version = byts[0]  # length of version string
    version = str(byts[1 : len_version + 1], "ascii")
    offset = 1 + len_version

    shape_info_nrows, shape_info_ncols = struct.unpack("<ii", byts[offset : offset + 8])
    shape_info_nbytes = shape_info_nrows * shape_info_ncols * itemsize
    offset += 8

    if shape_info_nbytes == 0:
        shape_info = np.array([], dtype=enc_dtype)
    else:
        shape_info = (
            np.frombuffer(byts[offset : offset + shape_info_nbytes], dtype=enc_dtype)
            .reshape(shape_info_nrows, shape_info_ncols)
            .copy()
        )
        offset += shape_info_nbytes

    byte_positions_rows = int.from_bytes(byts[offset : offset + 4], "little")
    byte_positions_nbytes = byte_positions_rows * 3 * itemsize
    offset += 4

    if byte_positions_nbytes == 0:
        byte_positions = np.array([], dtype=enc_dtype)
    else:
        byte_positions = (
            np.frombuffer(
                byts[offset : offset + byte_positions_nbytes], dtype=enc_dtype
            )
            .reshape(byte_positions_rows, 3)
            .copy()
        )
        offset += byte_positions_nbytes

    return version, shape_info, byte_positions, offset


def deserialize_chunk(
    byts: Union[bytes, memoryview], copy: bool = True, partial: bool = False
) -> Tuple[str, np.ndarray, np.ndarray, memoryview]:
    """Deserializes a chunk from the serialized byte stream. This is how the chunk can be accessed/modified after it is read from storage.

    Args:
        byts: (bytes) Serialized chunk.
        copy: (bool) If true, this function copies the byts while deserializing incase byts was a memoryview.
        partial: (bool) If true, the byts are only a part of the chunk.

    Returns:
        Tuple of:
        hub version used to create the chunk,
        encoded shapes info as numpy array,
        encoded byte positions as numpy array,
        chunk data as memoryview.

    Raises:
        IncompleteHeaderBytesError: For partial chunks, if the byts aren't complete to get the header.
    """
    incoming_mview = isinstance(byts, memoryview)
    byts = memoryview(byts)

    enc_dtype = np.dtype(hub.constants.ENCODING_DTYPE)
    itemsize = enc_dtype.itemsize

    # Read version
    len_version = byts[0]
    version = str(byts[1 : 1 + len_version], "ascii")
    offset = 1 + len_version

    # Read shape info
    shape_info_nrows, shape_info_ncols = struct.unpack("<ii", byts[offset : offset + 8])
    offset += 8
    shape_info_nbytes = shape_info_nrows * shape_info_ncols * itemsize
    if shape_info_nbytes == 0:
        shape_info = np.array([], dtype=enc_dtype)
    else:
        shape_info = (
            np.frombuffer(byts[offset : offset + shape_info_nbytes], dtype=enc_dtype)
            .reshape(shape_info_nrows, shape_info_ncols)
            .copy()
        )
        offset += shape_info_nbytes

    # Read byte positions
    byte_positions_rows = int.from_bytes(byts[offset : offset + 4], "little")
    offset += 4
    byte_positions_nbytes = byte_positions_rows * 3 * itemsize
    if byte_positions_nbytes == 0:
        byte_positions = np.array([], dtype=enc_dtype)
    else:
        byte_positions = (
            np.frombuffer(
                byts[offset : offset + byte_positions_nbytes], dtype=enc_dtype
            )
            .reshape(byte_positions_rows, 3)
            .copy()
        )
        offset += byte_positions_nbytes

    # Read data
    data = byts[offset:]
    if incoming_mview and copy:
        data = memoryview(bytes(data))
    return version, shape_info, byte_positions, data  # type: ignore


def serialize_chunkids(version: str, arr: np.ndarray) -> memoryview:
    """Serializes chunk ID encoders into a single byte stream. This is how the encoders will be written to the storage provider.

    Args:
        version: (str) Version of hub library.
        arr: (np.ndarray) Encoded chunk ids from a `ChunkIdEncoder` instance.

    Returns:
        Serialized chunk ids as memoryview.
    """
    len_version = len(version)
    flatbuff = bytearray(2 + len_version + arr.nbytes)

    # Write version
    len_version = len(version)
    flatbuff[0] = len_version
    flatbuff[1 : 1 + len_version] = version.encode("ascii")
    offset = 1 + len_version

    # write encoder dtype
    if version_compare(version, "2.7.6") >= 0:
        dtype = arr.dtype
        num_bytes = int(dtype.itemsize)
        flatbuff[offset] = num_bytes
        offset += 1

    # Write ids
    flatbuff[offset : offset + arr.nbytes] = arr.tobytes()
    offset += arr.nbytes
    return memoryview(flatbuff)


def deserialize_chunkids(
    byts: Union[bytes, memoryview]
) -> Tuple[str, np.ndarray, type]:
    """Deserializes a chunk ID encoder from the serialized byte stream. This is how the encoder can be accessed/modified after it is read from storage.

    Args:
        byts: (bytes) Serialized chunk ids.

    Returns:
        Tuple of: hub version used to create the chunk, encoded chunk ids as memoryview and dtype of the encoder.

    Raises:
        ValueError: If the bytes are not a valid chunk ID encoder.
    """
    byts = memoryview(byts)
    # Read version
    len_version = byts[0]
    version = str(byts[1 : 1 + len_version], "ascii")
    offset = 1 + len_version
    if version_compare(version, "2.7.6") < 0:
        # Read chunk ids
        ids = np.frombuffer(byts[offset:], dtype=np.uint32).reshape(-1, 2).copy()
        return version, ids, np.uint32
    else:
        # Read number of bytes per entry
        num_bytes = byts[offset]
        if num_bytes == 4:
            dtype = np.uint32
        elif num_bytes == 8:
            dtype = np.uint64  # type: ignore
        else:
            raise ValueError(f"Invalid number of bytes per entry: {num_bytes}")
        offset += 1
        # Read chunk ids
        ids = np.frombuffer(byts[offset:], dtype=dtype).reshape(-1, 2).copy()
        return version, ids, dtype


def serialize_sequence_or_creds_encoder(version: str, enc: np.ndarray) -> bytes:
    return len(version).to_bytes(1, "little") + version.encode("ascii") + enc.tobytes()


def deserialize_sequence_or_creds_encoder(
    byts: Union[bytes, memoryview], enc_type: str
) -> Tuple[str, np.ndarray]:
    dim = 2 if enc_type == "creds" else 3
    byts = memoryview(byts)
    len_version = byts[0]
    version = str(byts[1 : 1 + len_version], "ascii")
    enc = (
        np.frombuffer(byts[1 + len_version :], dtype=hub.constants.ENCODING_DTYPE)
        .reshape(-1, dim)
        .copy()
    )
    return version, enc


def check_sample_shape(shape, num_dims):
    if shape is not None and len(shape) != num_dims:
        raise TensorInvalidSampleShapeError(shape, num_dims)


def text_to_bytes(sample, dtype, htype):
    if isinstance(sample, hub.core.tensor.Tensor):
        try:
            if sample.htype == htype or sample.htype == "json" and htype == "list":
                return sample.tobytes(), sample.shape
        except (ValueError, NotImplementedError):  # sliced sample or tiled sample
            sample = sample.data()
    if htype in ("json", "list"):
        if isinstance(sample, np.ndarray):
            if htype == "list":
                sample = list(sample) if sample.dtype == object else sample.tolist()
            elif htype == "json":
                if sample.ndim == 0 or sample.dtype != object:
                    sample = sample.tolist()  # actually returns dict
                else:
                    sample = list(sample)
        validate_json_object(sample, dtype)
        byts = json.dumps(sample, cls=HubJsonEncoder).encode()
        shape = (len(sample),) if htype == "list" else (1,)
    else:  # htype == "text":
        if isinstance(sample, np.ndarray):
            sample = sample.tolist()
        if not isinstance(sample, str):
            raise TypeError("Expected str, received: " + str(sample))
        byts = sample.encode()
        shape = (1,)
    return byts, shape


def bytes_to_text(buffer, htype):
    buffer = bytes(buffer)
    if htype == "json":
        arr = np.empty(1, dtype=object)
        arr[0] = json.loads(bytes.decode(buffer), cls=HubJsonDecoder)
        return arr
    elif htype == "list":
        lst = json.loads(bytes.decode(buffer), cls=HubJsonDecoder)
        arr = np.empty(len(lst), dtype=object)
        arr[:] = lst
        return arr
    else:  # htype == "text":
        arr = np.array(bytes.decode(buffer)).reshape(
            1,
        )
    return arr


def serialize_text(
    incoming_sample: SampleValue,
    sample_compression: Optional[str],
    dtype: str,
    htype: str,
):
    """Converts the sample into bytes"""
    incoming_sample, shape = text_to_bytes(incoming_sample, dtype, htype)
    if sample_compression:
        incoming_sample = compress_bytes(incoming_sample, sample_compression)  # type: ignore
    return incoming_sample, shape


def serialize_numpy_and_base_types(
    incoming_sample: BaseTypes,
    sample_compression: Optional[str],
    chunk_compression: Optional[str],
    dtype: str,
    htype: str,
    min_chunk_size: int,
    break_into_tiles: bool = True,
    store_tiles: bool = False,
):
    """Converts the sample into bytes"""
    out = intelligent_cast(incoming_sample, dtype, htype)
    shape = out.shape
    tile_compression = chunk_compression or sample_compression

    if sample_compression is None:
        if out.nbytes > min_chunk_size and break_into_tiles:
            out = SampleTiles(out, tile_compression, min_chunk_size, store_tiles, htype)  # type: ignore
        else:
            out = out.tobytes()  # type: ignore
    else:
        ratio = get_compression_ratio(sample_compression)
        approx_compressed_size = out.nbytes * ratio

        if approx_compressed_size > min_chunk_size and break_into_tiles:
            out = SampleTiles(out, tile_compression, min_chunk_size, store_tiles, htype)  # type: ignore
        else:
            compressed_bytes = compress_array(out, sample_compression)
            out = compressed_bytes  # type: ignore

    return out, shape


def serialize_partial_sample_object(
    incoming_sample: PartialSample,
    sample_compression: Optional[str],
    chunk_compression: Optional[str],
    dtype: str,
    htype: str,
    min_chunk_size: int,
):
    shape = incoming_sample.shape
    return (
        SampleTiles(
            compression=sample_compression or chunk_compression,
            chunk_size=min_chunk_size,
            htype=htype,
            dtype=dtype,
            sample_shape=shape,
            tile_shape=incoming_sample.tile_shape,
        ),
        shape,
    )


def serialize_text_sample_object(
    incoming_sample: Sample, sample_compression: Optional[str]
):
    shape = incoming_sample.shape
    out = incoming_sample
    result = out.compressed_bytes(sample_compression)
    return result, shape


def serialize_sample_object(
    incoming_sample: Sample,
    sample_compression: Optional[str],
    chunk_compression: Optional[str],
    dtype: str,
    htype: str,
    min_chunk_size: int,
    break_into_tiles: bool = True,
    store_tiles: bool = False,
):
    shape = incoming_sample.shape
    tile_compression = chunk_compression or sample_compression

    out = incoming_sample
    if sample_compression:
        compression_type = get_compression_type(sample_compression)
        is_byte_compression = compression_type == BYTE_COMPRESSION
        if is_byte_compression and out.dtype != dtype:
            # Byte compressions don't store dtype, need to cast to expected dtype
            arr = intelligent_cast(out.array, dtype, htype)
            out = Sample(array=arr)

        compressed_bytes = out.compressed_bytes(sample_compression)

        if (
            compression_type not in (VIDEO_COMPRESSION, AUDIO_COMPRESSION)
            and len(compressed_bytes) > min_chunk_size
            and break_into_tiles
        ):
            out = SampleTiles(  # type: ignore
                out.array, tile_compression, min_chunk_size, store_tiles, htype
            )
        else:
            out = compressed_bytes  # type: ignore
    else:
        out = intelligent_cast(out.array, dtype, htype)  # type: ignore

        if out.nbytes > min_chunk_size and break_into_tiles:  # type: ignore
            out = SampleTiles(out, tile_compression, min_chunk_size, store_tiles, htype)  # type: ignore
        else:
            out = out.tobytes()  # type: ignore
    return out, shape


def serialize_tensor(
    incoming_sample: "hub.core.tensor.Tensor",
    sample_compression: Optional[str],
    chunk_compression: Optional[str],
    dtype: str,
    htype: str,
    min_chunk_size: int,
    break_into_tiles: bool = True,
    store_tiles: bool = False,
):
    def _return_numpy():
        return serialize_numpy_and_base_types(
            incoming_sample.numpy(),
            sample_compression,
            chunk_compression,
            dtype,
            htype,
            min_chunk_size,
            break_into_tiles,
            store_tiles,
        )

    if incoming_sample.meta.chunk_compression or chunk_compression:
        return _return_numpy()
    elif incoming_sample.meta.sample_compression == sample_compression:
        # Pass through
        try:
            return incoming_sample.tobytes(), incoming_sample.shape  # type: ignore
        except (
            ValueError,
            NotImplementedError,
        ) as e:  # Slice of sample or tiled sample
            return _return_numpy()
    else:
        return _return_numpy()



from typing import Callable, Tuple
import numpy as np


def break_into_tiles(sample: np.ndarray, tile_shape: Tuple[int, ...]) -> np.ndarray:
    """Get a new tile-ordered numpy object array that is the shape of the tile grid.
    Each element of the returned numpy object array is also a numpy array that is a zero-copy view of the actual tile
    at the tile coordinate.

    Example:
        >>> tiles = break_into_tiles(np.arange(10).reshape(2, 5), (3, 3))
        >>> tiles
        [[array([[0, 1, 2],
                [5, 6, 7]])
        array([[3, 4],
                [8, 9]])]]
        >>> tiles.shape
        (1, 2)
        >>> tiles[0, 0]
        array([[0, 1, 2],
            [5, 6, 7]])
        >>> tiles[0, 1]
        array([[3, 4],
            [8, 9]])

    Args:
        sample: The sample array to break into tiles.
        tile_shape: The shape that each tile will have. Determines the number of tiles in each dimension.
            Corner tiles (tiles at the end of one or more dimensions) may be smaller than the tile shape.

    Returns:
        numpy object array of tiles. Each element of the array is a numpy array that is a zero-copy view (source = sample array)
            of the actual tile at the tile coordinate.
    """

    tiles_per_dim = np.ceil(np.divide(sample.shape, tile_shape)).astype(int)
    tiles = np.empty(tiles_per_dim, dtype=object)
    for tile_coords, _ in np.ndenumerate(tiles):
        low = np.multiply(tile_coords, tile_shape)
        high = low + tile_shape
        idx = tuple(slice(l, h) for l, h in zip(low, high))
        tiles[tile_coords] = sample[idx]
    return tiles


def get_tile_shapes(sample_shape: Tuple[int, ...], tile_shape: Tuple[int, ...]):
    tiles_per_dim = np.ceil(np.divide(sample_shape, tile_shape)).astype(int)
    tile_shapes = np.empty(tiles_per_dim, dtype=object)
    for tile_coords in np.ndindex(*tiles_per_dim):
        low = np.multiply(tile_coords, tile_shape)
        high = np.minimum(low + tile_shape, sample_shape)
        tile_shapes[tile_coords] = tuple(high - low)
    return tile_shapes


def serialize_tiles(
    tiles: np.ndarray, serialize_func: Callable[[np.ndarray], bytes]
) -> np.ndarray:
    """Get a new tile-ordered numpy object array that is the same shape of the tile-grid.
    Each element of the returned numpy object array is a memoryview object representing the serialized tile.

    Args:
        tiles (np.ndarray): The tile-ordered numpy object array to serialize.
        serialize_func (Callable[[np.ndarray], memoryview]): A function that takes a numpy array and returns a memoryview object.
            This function is used to serialize each tile, may be used to compress the tile.

    Returns:
        np.ndarray: numpy object array of serialized tiles. Each element of the array is a memoryview object.
    """
    return np.vectorize(serialize_func, otypes=[object])(tiles)



from typing import Tuple, Optional, Union
import numpy as np


class PartialSample:
    """Represents a sample that is initialized by just shape and the data is updated later."""

    def __init__(
        self,
        sample_shape: Tuple[int, ...],
        tile_shape: Optional[Tuple[int, ...]] = None,
        dtype: Union[str, np.dtype] = np.dtype("uint8"),
    ):
        self.sample_shape = sample_shape
        self.tile_shape = tile_shape
        self.dtype = dtype

    @property
    def shape(self):
        return self.sample_shape

    def astype(self, dtype: Union[str, np.dtype]):
        return self.__class__(self.sample_shape, self.tile_shape, dtype)
